"""
This section covers the topic of logging during execution of the Deltaflow
graph and contains the following user-facing parts:

- Factory :py:func:`make_logger` acts as an interface to the Deltaflow
  logging. We present some use cases in
  `Logging and tracing within Deltaflow <tutorials/logging.html>`_.

For convenience the classes and functions listed above are also exposed to
the top level module and can be used like this:

.. code-block:: python

    import deltalanguage as dl
    
    logger = dl.make_logger(30, 'test_logger')
    logger.warning("We have a warning!")
"""

import logging
from typing import Dict

from deltalanguage._utils import QueueMessage


# Global dictionary with all deltalanguage loggers
DeltaLoggers: Dict[str, logging.Logger] = {}
DeltaHandlers: Dict[str, logging.Handler] = {}


def make_logger(lvl: int, name: str) -> logging.Logger:
    """Produce a logger with sensible properties.

    Parameters
    ----------
    lvl : int
        Logging level.
    name : str
        Logger name.

    Examples
    --------
    Only messages above the defined level will be logged:

    .. code-block:: python

        log = make_logger(logging.WARNING, "Test")
        log.error("Error example")
        # Test [ERROR]: Error example
        log.warning("Warning example")
        # Test [WARNING]: Warning example
        log.info("Info example")
        log.debug("Debug example")
    """
    global DeltaLoggers

    if name in DeltaLoggers:
        return DeltaLoggers[name]

    log = logging.getLogger(name)
    log.setLevel(lvl)
    if name in DeltaHandlers:
        # log already has handler, we just need to set the level
        DeltaHandlers[name].setLevel(lvl)
    else:
        ch = logging.StreamHandler()
        ch.setLevel(lvl)

        fmt = logging.Formatter("%(name)s [%(levelname)s]: %(message)s")

        ch.setFormatter(fmt)
        log.addHandler(ch)
        DeltaHandlers[name] = ch

    DeltaLoggers[name] = log
    return log


def get_handler(name: str) -> logging.Handler:
    """Helper function that returns the logging handler for the log of the
    given name.

    Required for updating the level of logging in a node.

    Parameters
    ----------
    name : str
        The name the log is recorded under.

    Returns
    -------
    logging.Handler
        The handler for the given log.
    """
    return DeltaHandlers[name]


def clear_loggers():
    """Resets the logs after program has terminated."""
    global DeltaLoggers
    DeltaLoggers = {}


class MessageLog:
    """Record of messages sent during graph simulation.

    In order to track messages, we associate a logical clock with each
    node and each message. This starts at 0 and gets updated in 2 cases:

    - When an input is received the node's clock is reset to the maximum
      clock value of all inputs and the internal clock.

    - When an output is sent the node's clock is incremented by 1 and
      the output message gets assigned the same value.
      This ensures that the internal state of the node is taken in account.

    Parameters
    ----------
    lvl : int
        Logging level.

    Attributes
    ----------
    messages : List[QueueMessage]
        Log of messages.


    .. todo :: This class belongs to the domain of `DeltaPySimulator`.
    """

    def __init__(self, lvl: int = logging.ERROR):
        self.messages = []
        self.lvl = lvl

    def add_message(self, sender: str, port: str, msg: QueueMessage):
        """Add a message to the log.

        Note, only messages with non-negative clock are added.
        Negative clock messages are used for flushing.

        Examples
        --------

        All messages are logged at :py:const:`logging.INFO`.
        To view message logs, use the `msg_lvl` parameter of
        :py:class:`DeltaPySimulator<deltalanguage.runtime.DeltaPySimulator>`.

        .. code-block:: python

            with DeltaGraph() as graph:
                print_then_exit(add(2, 3))
            rt = DeltaPySimulator(graph, msg_lvl=logging.INFO)
            rt.run()
            # MessageLog [INFO]: Sent to print_then_exit_4 port val at logical clock time 2: 5
            # 5
        """

        if not isinstance(msg, QueueMessage):
            raise TypeError("Message logged that was not of"
                            + f"type QueueMessage: {msg}")
        if (self.lvl <= logging.INFO
                and msg.msg is not None
                and msg.clk >= 0):
            self.messages.append((sender, port, msg))

    def log_messages(self):
        """Print all messages in order of clock time."""
        self.messages.sort(key=lambda x: x[2].clk)
        log = make_logger(self.lvl, "MessageLog")
        for sender, port, msg in self.messages:
            log.info(f"Sent to {sender} port {port} at logical "
                     + f"clock time {msg.clk}: {msg.msg}")



"""
In order to facilitate data exchange between nodes written in potentially
different languages we use these intermediate data types to encode
to/from the data types used inside of the nodes.

This submodule contains the following user-facing parts:

- Primitive data types:
  :py:class:`DBool`
  :py:class:`DChar`
  :py:class:`DInt`
  :py:class:`DUInt`
  :py:class:`DFloat`
  :py:class:`DComplex`

- Compound data types:
  :py:class:`DArray`
  :py:class:`DStr`
  :py:class:`DTuple`
  :py:class:`DRecord`

- Special data types:
  :py:class:`Top`
  :py:class:`DRaw`
  :py:class:`DUnion`
  :py:class:`DSize`
  :py:class:`DOptional`
  :py:class:`Void`

- Auxiliary functions:
  :py:func:`make_forked_return`
  :py:func:`as_delta_type`
  :py:func:`delta_type`

For convenience the classes and functions listed above are also exposed to
the top level module and can be used like this:

.. code-block:: python

    import deltalanguage as dl
    
    print(dl.DBool())
    print(dl.delta_type(5))

More information about this submodule as well as examples of use can be
found in `Deltaflow Data Types <tutorials/data_types.html>`_.

"""

from ._delta_types import (BaseDeltaType,
                           PrimitiveDeltaType,
                           CompoundDeltaType,
                           Top,
                           DBool,
                           DChar,
                           DComplex,
                           DFloat,
                           DInt,
                           DArray,
                           DOptional,
                           DRaw,
                           DRecord,
                           DStr,
                           DTuple,
                           DUInt,
                           DUnion,
                           ForkedReturn,
                           as_delta_type,
                           delta_type,
                           make_forked_return)
from ._exceptions import DeltaIOError, DeltaTypeError
                          
from ._special import DSize, Void

# user-facing classes
__all__ = ["Top",
           "DBool",
           "DChar",
           "DInt",
           "DUInt",
           "DFloat",
           "DComplex",
           "DArray",
           "DStr",
           "DTuple",
           "DRecord",
           "DRaw",
           "Top",
           "DSize",
           "DUnion",
           "DOptional",
           "Void",
           "make_forked_return",
           "as_delta_type",
           "delta_type"]



"""Module defining Real Node types, which can be run using our DeltaPySimulator."""
from __future__ import annotations
from inspect import _empty, signature
import logging
from queue import Full, Queue
import sys
import textwrap
from threading import Event
from time import sleep

from typing import (TYPE_CHECKING,
                    Any,
                    Callable,
                    Dict,
                    List,
                    Optional,
                    OrderedDict,
                    Tuple,
                    Type,
                    Union)

import dill

from deltalanguage.data_types import (BaseDeltaType,
                                      DeltaTypeError,
                                      DOptional,
                                      ForkedReturn,
                                      Void,
                                      DeltaIOError,
                                      as_delta_type,
                                      delta_type)
from deltalanguage.logging import MessageLog, make_logger
from deltalanguage._utils import (NamespacedName,
                                  QueueMessage)

from .._body_templates import BodyTemplate, MethodBodyTemplate
from .abstract_node import AbstractNode, ForkedNode
from .node_bodies import (PyConstBody,
                          PyFuncBody,
                          PyMethodBody,
                          PyInteractiveBody,
                          PyMigenBody,
                          Body)
from .port_classes import InPort, OutPort

if TYPE_CHECKING:
    from deltalanguage.runtime import DeltaQueue, DeltaPySimulator
    from .. import DeltaGraph


class RealNode(AbstractNode):
    """Class to represent a non-abstract node that can form part of
    :py:class:`DeltaGraph`.

    Parameters
    ----------
    graph : DeltaGraph
        Graph this node is a member of.
    bodies : List[Body]
        List of bodies that can represent the implementation for this node.
    out_type : BaseDeltaType
        The type of what we expect the body to output.
    name : Optional[str]
        The name of the node (an index is appended to the end).
    lvl : int
        Logging level for the node.
        By default only error logs are displayed.
    is_autogenerated : bool
        True if this node is created automatically to provide an input to
        another node. For instance:

        .. code-block:: python

            with dl.DeltaGraph() as graph:
                printer(42)

        has an autogenerated node that provides 42. For strict typing this
        node should send data in the same format as printer's input.
    """

    def __init__(self,
                 graph,
                 bodies: List[Body],
                 in_params: Dict[str, BaseDeltaType] = {},
                 out_type: BaseDeltaType = None,
                 name: Optional[str] = None,
                 lvl: int = logging.ERROR,
                 is_autogenerated: bool = False):
        self.graph = graph
        self.graph.add_node(self)  # Registering self with parent graph

        self.is_autogenerated = is_autogenerated

        self._body = None
        self.bodies = bodies

        if len(self.bodies) == 1:
            self._body = bodies[0]

        self.in_params = in_params
        self.out_type = out_type

        idx = RealNode.get_next_index()
        if name is None:
            # set my name to the next unique available name
            self._name = f"node_{idx}"
        else:
            self._name = f"{name}_{idx}"

        # Ports in/out to this node
        self.in_ports: Dict[NamespacedName, InPort] = {}
        self.out_ports: List[OutPort] = []

        self.log = make_logger(lvl,
                               f"{self.__class__.__name__} {self._name}")

        # See MessageLog for detail
        self._clock = 0

        self.fork_names = None
        if isinstance(self.out_type, ForkedReturn):
            self.fork_names = self.out_type.keys

            for fork_name in self.fork_names:
                if fork_name in dir(self):
                    raise NameError("Invalid fork name: " +
                                    fork_name + " for node " + self.name)

    def select_body(self,
                    exclusions: List[object] = [],
                    preferred: List[object] = [],
                    override=True):
        """Select a body for this node, out of the existing list of bodies
        attached to it.
        Exclusions take priority over preferences. If multiple choices
        remain the earliest added node will be selected.

        Parameters
        ----------
        exclusions : List[object]
            List of keys to exclude from selection
        preferred : List[object]
            List of keys to be preferred for selection if available
        override : Bool
            If true, then this selection will override existing selection
            Otherwise this selection will not occur if there has already
            been a selection

        Returns
        -------
        bool
            ``True`` if a ``body`` is defined at the end of selection.

        """

        if override or not self._body:

            not_excluded_list = []
            for body in self.bodies:
                # If body has no excluded tag
                if not [tag for tag in exclusions if tag in body.access_tags]:
                    not_excluded_list.append(body)

            if not_excluded_list:
                prefered_list = []
                for body in not_excluded_list:
                    # If body has a preferred tag
                    if [tag for tag in preferred if tag in body.access_tags]:
                        prefered_list.append(body)

                if prefered_list:
                    self._body = prefered_list[0]
                else:
                    self._body = not_excluded_list[0]
            else:
                self._body = None
        return bool(self._body)

    def __str__(self) -> str:

        indent_prefix = "    "

        ret = f'ports:\n'
        if self.in_ports:
            ret += textwrap.indent(
                f'in:\n' + textwrap.indent(
                    ''.join(
                        [f"{in_port}\n" for in_port in self.in_ports.values()]
                    ),
                    prefix=indent_prefix),
                prefix=indent_prefix)
        if self.out_ports:
            ret += textwrap.indent(
                f'out:\n' + textwrap.indent(
                    ''.join(
                        [f"{out_port}\n" for out_port in self.out_ports]
                    ),
                    prefix=indent_prefix),
                prefix=indent_prefix)

        ret += f'bodies:\n'
        if not self.bodies:
            ret += textwrap.indent('None\n\n', prefix=indent_prefix)
        else:
            ret += textwrap.indent(
                ''.join(
                    [
                        f'{"*" if body is self._body else ""}' +
                        f'{body.__class__.__name__}\n' +
                        textwrap.indent(
                            f'tags: {next(i for i in body.access_tags if i is not body.__class__)}\n',
                            prefix=indent_prefix
                        ) for body in self.bodies
                    ]
                ),
                prefix=indent_prefix
            ) + '\n'

        ret = f"node[{self.name}]:\n" + textwrap.indent(
            ret, prefix=indent_prefix
        )

        return ret

    def __repr__(self):
        return self.name

    def __getattr__(self, item):
        if self.fork_names and item in self.fork_names:
            return ForkedNode(self, item)
        elif self.fork_names is None:
            raise AttributeError(
                f"Cannot fetch {item} from {self.name}, as we don't "
                f"have multiple outputs. Suggest using the node on it's own."
            )
        elif item == '__del__':
            raise AttributeError('Migen tries to get this attribute')
        else:
            raise DeltaIOError(
                f"Node {self.name} has out ports {list(self.fork_names)}. "
                f"The requested out port \'{item}\' is not found. "
                "Please check the node's definition."
            )

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    def add_out_port(self, port_destination: InPort, index=None):
        """Creates an out-port and adds it to my out-port store.

        Parameters
        ----------
        index : Optional[str]
            If the out-port is one of several for this node, index specifies
            what part of the output is sent via this port. If the node has
            only one output, then this is `None`.
        port_destination : InPort
            The in-port that this out-port exports to.
        """
        try:
            if issubclass(self.out_type, Void):
                raise DeltaIOError(
                    f"Cannot make an out-port on node {self.name} "
                    "with return type \'Void\'.\n"
                    "Please either add the proper return type to the "
                    "node definition or do not try to create an "
                    "output data channel from this node."
                )
        except TypeError:
            pass

        if index is None:  # out-port type is whole node return type
            # due to the strict typing the out type should match the in type
            if self.is_autogenerated:
                type_out = port_destination.port_type
            else:
                type_out = as_delta_type(self.out_type)

            self.out_ports.append(
                OutPort(NamespacedName(self.name, None),
                        type_out,
                        port_destination, self)
            )
        else:  # out-port type is indexed node return type
            self.out_ports.append(
                OutPort(NamespacedName(self.name, index),
                        as_delta_type(self.out_type.elem_dict[index]),
                        port_destination, self)
            )

        # If this port is going into a port on a different graph,
        # flatten this graph into said graph
        into_graph = port_destination.node.graph
        if into_graph is not self.graph:
            into_graph.flatten(self.graph)

    def add_in_port(self, arg_name: str, in_type: Type, in_port_size: int = 0):
        """Creates an in-port and adds it to my in-port store.

        Parameters
        ----------
        arg_name : str
            Name of the argument this port supplies.
        in_type : Type
            The type that is expected to be supplied for this port.
        in_port_size: int
            Maximum size of the in ports.
            If 0 then size is unlimited.

        Returns
        -------
        InPort
            The created port.
        """
        my_port = InPort(NamespacedName(self.name, arg_name),
                         as_delta_type(in_type),
                         self,
                         in_port_size)
        self.in_ports[my_port.port_name] = my_port
        return my_port

    def _create_upstream_ports(self,
                               required_in_ports: Dict[str, Type],
                               given_nodes: Dict[str, AbstractNode],
                               in_port_size: int = 0):
        """Create the ports going into this node and their
        corresponding out-ports.

        Parameters
        ----------
        required_in_ports
            Dictionary that describes what in-ports we want coming in to
            this node.
        given_nodes
            The nodes that are expected to send results to out in-ports.
        in_port_size
            The maximum size of the node's in ports. If 0 then unlimited size.
        """
        for arg_name, type_wanted in required_in_ports.items():
            if not arg_name in given_nodes:
                raise DeltaIOError(
                    f"Node {self.name} has a mismatch between defined and "
                    f"used input ports.\nArgument \'{arg_name}\' cannot be "
                    f"found in given ports {list(given_nodes.keys())}."
                )

            in_port = self.add_in_port(arg_name, type_wanted, in_port_size)
            given_nodes[arg_name].add_out_port(in_port, None)

    def _ports_from_arguments(self,
                              required_in_params,
                              pos_in_nodes,
                              kw_in_nodes,
                              in_port_size=0):
        """Manages the creation of upstream ports by getting the given nodes
        into the correct data structure.

        Parameters
        ----------
        required_in_params
            Dictionary that describes what in-ports we want coming in to
            this node.
        pos_in_nodes
            Nodes expected to send results to this node, specified
            positionally.
        kw_in_nodes
            Nodes expected to send results to this node, specified by keyword.
        in_port_size
            The maximum size of the node's in ports. If 0 then unlimited size.
        """
        if not isinstance(required_in_params, OrderedDict) and pos_in_nodes:
            raise TypeError("Please specify input parameter types to blocks "
                            "via an OrderedDict if you want to specify "
                            "parameters by position")

        # wrap the given positional in_nodes up with the name of the param they
        # satisfy
        pos_in_nodes_dict = {
            param_name: given_node
            for (given_node, (param_name, _)) in zip(
                pos_in_nodes,
                required_in_params.items()
            )
        }

        self._create_upstream_ports(required_in_params,
                                    {**pos_in_nodes_dict, **kw_in_nodes},
                                    in_port_size=in_port_size)

    @property
    def name(self) -> str:
        return self._name

    @property
    def body(self):
        if self._body:
            return self._body
        else:
            if self.bodies:
                raise ValueError("Please call select_body on node "
                                 f"{self.name} before body access")
            else:
                raise ValueError("No bodies available for selection on node "
                                 f"{self.name}.")


class PythonNode(RealNode):
    """Parent Node type for all Python constructs.

    Attributes
    ----------
    in_queues : Dict[str, DeltaQueue]
        Queues providing input(s).
    out_queues : Dict[str, DeltaQueue]
        Queues consumins output(s).
    sig_stop : Event
        Communication channel through which a runtime simulator or a runtime
        signals `thread_worker` to stop.
    node_key : Optional[str]
        Keyword argument used for providing the node to the block, included for
        debugging purposes.
    lvl : int
        Logging level for the node.
        By default only error logs are displayed.
    """

    def __init__(self,
                 graph,
                 bodies,
                 in_params,
                 pos_in_nodes,
                 kw_in_nodes,
                 in_port_size: int = 0,
                 node_key: Optional[str] = None,
                 out_type=None,
                 name: str = None,
                 lvl: int = logging.ERROR,
                 is_autogenerated: bool = False):
        super().__init__(graph=graph,
                         bodies=bodies,
                         in_params=in_params,
                         out_type=out_type,
                         name=name,
                         lvl=lvl,
                         is_autogenerated=is_autogenerated)

        self._is_const = False

        const_body_no = sum(isinstance(body, PyConstBody) for body in bodies)
        if const_body_no > 0:
            if const_body_no == len(bodies):
                self._is_const = True
                for in_type in self.in_params.values():
                    if isinstance(in_type, DOptional):
                        raise TypeError('Optional input is not allowed '
                                        'for constant nodes')
                    elif not isinstance(in_type, BaseDeltaType):
                        raise TypeError('Unsupported data type')
            else:
                raise ValueError("For a multi-body node, if one body is const "
                                 "all bodies must be const")

        self.in_queues: Dict[str, DeltaQueue] = None
        self.out_queues: Dict[str, DeltaQueue] = None
        self.sig_stop: Event = None
        self.node_key = node_key
        self.in_port_size = in_port_size

        self._ports_from_arguments(self.in_params, pos_in_nodes,
                                   kw_in_nodes, self.in_port_size)

    def set_communications(self, runtime: DeltaPySimulator):
        """Get the in and out queues relating to this node, as well as the
        utility events such as ``sig_stop`` from a runtime simulator or
        a runtime and save them in the instance.

        Parameters
        ----------
        runtime : DeltaPySimulator
            API of a runtime simulator or a runtime.
        """
        self.in_queues = runtime.in_queues[self.name]
        self.out_queues = runtime.out_queues[self.name]
        self.sig_stop = runtime.sig_stop

    def check_stop(self):
        """Check the stop signal, which can be set by a runtime simulator or
        a runtime. If set, it stops the current thread.
        """
        if self.sig_stop.is_set():
            self.log.info(f"Stopped {repr(self)}.")
            sys.exit()

    def receive(self, *args: str) -> Union[Dict[str, Any], Any]:
        """Receives the node's input(s) via in ports.

        Compulsory inputs block the further execution until the data is
        provided, whereas optional inputs do not block.

        Check if the node should stop.

        Parameters
        ----------
        args : str
            Optionally filter inputs. Only the specified ones will be received.

        Returns
        -------
        Union[Dict[str, Any], Any]
            If there is one input is specified via ``args`` it is received as
            an object, overwise the input values are returned as a dictionary.
        """
        # receive all or only selected inputs
        if args:
            in_queues = {name: in_q
                         for name, in_q in self.in_queues.items()
                         if name in args}
        else:
            in_queues = self.in_queues

        values = {}
        for name, in_q in in_queues.items():
            values[name] = in_q.get()
            if not isinstance(values[name], QueueMessage):
                raise TypeError(
                    f"Queue {in_q} from port {repr(in_q._src)} contained "
                    f"an item which was not a QueueMessage: {values[name]}"
                )
            self.msg_log.add_message(self.name, name, values[name])

        # logical clock update
        self._clock = max([self._clock] + [v.clk for v in values.values()])

        # unpack the inner msg
        val = {k: v.msg for k, v in values.items()}

        self.check_stop()

        if all(v == None for v in val.values()):
            # let the Python GIL take a look at the other threads
            sleep(1e-9)

        # if there is just one value to return, unpack it from the dict
        if len(val) == 1 and args:
            val = list(val.values())[0]

        if val:
            self.log.info(f"<- {val}")

        return val

    def send(self, ret: Union[object, Tuple]):
        """Sends the node's output(s) via out ports.

        If sending is blocked this method will lock the execution of the node
        until unblocked or a stop signal is raised, which shall stop
        the execution of the node.

        Parameters
        ----------
        ret : Union[object, Tuple]
            The return value. It is implied by construction that if it is a
            single object then the node has only one out port, otherwise
            a named tuple is used, with the names of the fields matching
            the names of the out ports.
        """
        # Log only non-trivial output(s)
        if not ((isinstance(ret, Tuple) and all(map(lambda x: x is None, ret)))
                or ret is None):
            self.log.info(f"-> {ret}")

        self._clock += 1
        message = QueueMessage(ret, clk=self._clock)

        for out_q in self.out_queues.values():
            while True:
                try:
                    out_q.put(message)
                    self.check_stop()
                    break
                except Full:
                    self.check_stop()
                except:
                    raise

        # let the Python GIL take a look at the other threads
        sleep(1e-9)

    def thread_worker(self, runtime: DeltaPySimulator):
        """Run a regular Python node.

        Waits for input on all the mandatory inputs.
        Then, de-queues the optional inputs and executes its body.
        The output is written to the appropriate output queues.

        Parameters
        ----------
        runtime : DeltaPySimulator
            API of a runtime simulator or a runtime.
        """
        if isinstance(self.body, PyConstBody):
            # Thread worker should never run on a const body
            raise NotImplementedError

        if isinstance(self.body, PyInteractiveBody):
            self.log.debug(f"Running... {self}")
            self.body.eval(self)

        else:
            while True:
                values = self.receive()

                # If a node keyword has been specified for debugging then add
                # the node to the arguments.
                if self.node_key:
                    # the self arg is effectively a const message,
                    # so from time 0
                    values[self.node_key] = self

                self.log.debug("Running...")
                ret = self.body.eval(**values)
                self.send(ret)

    def run_once(self, runtime: DeltaPySimulator):
        """Compute the value of the node and pass it to the output queues.

        The output queues are
        :py:class:`ConstQueue<deltalanguage.runtime.ConstQueue>` that store
        this value and retrive a deepcopy at each request of the receiving
        node.
        This is done purely for optimisation purposes.

        Parameters
        ----------
        runtime : DeltaPySimulator
            API of a runtime simulator or a runtime.
        """
        ret = self.body.eval()
        self.send(ret)

    def capnp(self, capnp_node, capnp_bodies):
        """Generate ``capnp`` form of this node.

        Parameters
        ----------
        capnp_node
            The capnp object of this node.
        capnp_bodies
            List of bodies so we can check if a body is already serialised.
        """
        capnp_node.name = self.name
        capnp_node.init("bodies", len(self.bodies))

        for i_bod, bod in enumerate(self.bodies):

            body_impl = bod.as_serialized

            if isinstance(bod, PyMigenBody):
                body_id = 'migen'
                impl_type = 'verilog'
            elif isinstance(bod, PyInteractiveBody):
                body_id = 'interactive'
                impl_type = 'dillImpl'
            else:
                body_id = 'python'
                impl_type = 'dillImpl'

            def get_body_impl(body):
                return body.__getattr__(body_id).__getattr__(impl_type)

            def set_body_impl(body, impl):
                body.__getattr__(body_id).__setattr__(impl_type, impl)

            for i, body in enumerate(capnp_bodies):
                if body.which() == body_id:
                    if get_body_impl(body) == body_impl:
                        capnp_node.bodies[i_bod] = i
                        break
            else:
                body = capnp_bodies.add()
                body.init(body_id)
                set_body_impl(body, body_impl)
                capnp_node.bodies[i_bod] = len(capnp_bodies) - 1

        # 2. save I/O ports
        self.capnp_ports(capnp_node)

    def capnp_ports(self, capnp_node):
        """Helper method, generates capnp for in/out ports of the node.

        Parameters
        ----------
        capnp_node
            The node of the interest.
        """
        in_ports = capnp_node.init("inPorts", len(self.in_ports))
        for capnp_in_port, in_port in zip(in_ports, self.in_ports.values()):
            in_port.capnp(capnp_in_port)

        out_ports = capnp_node.init("outPorts", len(self.out_ports))
        for capnp_out_port, out_port in zip(out_ports, self.out_ports):
            out_port.capnp(capnp_out_port)

    def capnp_wiring(self, capnp_nodes, capnp_wiring):
        """Generate capnp form of this node's wires.

        Parameters
        ----------
        capnp_nodes
            List of nodes so indexes can be found.
        capnp_wiring
            List of wires so we can add our relevant wires.
        """
        for i, capnp_node in enumerate(capnp_nodes):
            if capnp_node.name == self.name:
                capnp_node_index = i
                break

        for i, out_port in enumerate(self.out_ports):
            capnp_wire = capnp_wiring.add()
            capnp_wire.srcNode = capnp_node_index
            capnp_wire.srcOutPort = i
            out_port.capnp_wiring(capnp_nodes, capnp_wire)

    def set_msg_log(self, msg_log: MessageLog):
        """Sets the log for messages received.

        Parameters
        ----------
        msg_log : MessageLog
            Instance of the message log.
        """
        self.msg_log = msg_log

    def add_body(self, body: Union[BodyTemplate, Callable]):
        """Add some body to this node, using a ``BodyTemplate`` as a recipe
        for how to construct the body to add. This will also enable checks
        about body compatibility to be made.

        Parameters
        ----------
        body : Union[BodyTemplate, Callable]
            ``BodyTemplate`` to add or some object for which
            ``template`` is defined, where ``template`` is a ``BodyTemplate``
        """
        if self.is_const():
            raise NotImplementedError("Adding a body to a const node "
                                      "post-creation is currently unsupported")

        template = body
        if not isinstance(template, BodyTemplate):
            template = body.template

        if template.compatible(self.node_key, self.in_port_size,
                               self.in_params, self.out_type):
            if isinstance(template, MethodBodyTemplate):
                self.bodies.append(template.construct_body(body.__self__))
            else:
                self.bodies.append(template.construct_body())
            self.select_body(override=False)

    def is_const(self):
        """Method to return state of self._is_const.

        Returns
        -------
        bool
            If the current body is constant or not.
        """
        return self._is_const


def as_node(potential_node: Union[AbstractNode, object],
            graph: DeltaGraph) -> PythonNode:
    """Ensures argument is a node and if not makes it into a constant node.

    Parameters
    ----------
    potential_node : Union[AbstractNode, object]
        Node that could be a node or not.
    graph : DeltaGraph
        Graph the node would be in

    Returns
    -------
    PythonNode
        Made for potential_node or potential_node as it was already a node.
    """
    if isinstance(potential_node, AbstractNode):
        return potential_node
    else:
        return PythonNode(graph,
                          [PyConstBody(lambda: potential_node)],
                          {},
                          [],
                          {},
                          out_type=delta_type(potential_node),
                          is_autogenerated=True)


def get_func_in_params_out_type(
    a_func: Callable,
    is_method: bool,
    node_key: Optional[str] = None
) -> Tuple[OrderedDict[str, Union[BaseDeltaType, DOptional]],
           Union[BaseDeltaType, ForkedReturn]]:
    """Helper function to extract input and output types of a node function.

    Parameters
    ----------
    a_func : Callable
        The function to analyse.
    is_method : bool
        Flag to specify if function is a class method.
    node_key : Optional[str]
        Keyword argument used for providing the node to the block, included for
        some logic purposes.

    Returns
    -------
    OrderedDict[str, Union[BaseDeltaType, DOptional]]
        Types of the in parameters.
    Union[BaseDeltaType, ForkedReturn]
        Type of the output the node to be made.

    Raises
    ------
    TypeError
        Raised if either the input or output types aren't specified in the
        function signature.
    """
    func_args = signature(a_func).parameters
    out_type = signature(a_func).return_annotation

    in_params = OrderedDict()
    for i, (arg_name, arg_param) in enumerate(func_args.items()):

        # first argument should always be 'self' for a method
        if i == 0 and is_method:
            continue

        if arg_param.annotation == _empty:
            raise TypeError(
                "Must specify the type of argument " +
                f"'{arg_name}' as annotation in " +
                f"function '{a_func.__name__}'"
            )
        in_params[arg_name] = arg_param.annotation

    in_params = in_params_as_delta_types(in_params, node_key)

    if out_type == _empty:
        out_type = Void

    delta_type_out = as_delta_type(out_type)
    if delta_type_out is not Void:
        if not isinstance(delta_type_out, (BaseDeltaType, ForkedReturn)):
            raise DeltaTypeError(f"Unsupported type={out_type}")

    return in_params, delta_type_out


def in_params_as_delta_types(in_params: Dict[str, Type],
                             node_key: Optional[str] = None
                             ) -> Dict[str, Union[BaseDeltaType, DOptional]]:
    """Take ``in_params`` and convert to DeltaTypes, raising appropriate
    errors, skips and removes ``node_key``.

    Parameters
    ----------
    in_params : Dict[str, Type]
        Dictionary of in_params that need to become DeltaTypes
    node_key : Optional[str], optional
        Node key for node param if available, by default None

    Returns
    -------
    Dict[str, Union[BaseDeltaType, DOptional]]
        Dictionary of in_params as DeltaTypes
    """
    for param_name, param_type in in_params.items():

        if node_key == param_name and param_type == PythonNode:
            in_params.pop(node_key)
            continue

        delta_type_in = as_delta_type(param_type)
        if not isinstance(delta_type_in, (BaseDeltaType, DOptional)):
            raise DeltaTypeError(f"Unsupported type={param_type}")

        in_params[param_name] = delta_type_in
    return in_params



import atexit

import numpy as np
from numpy.random import RandomState

from projectq import MainEngine
from projectq.backends import CircuitDrawer, Simulator
from projectq.ops import (All, C, DaggeredGate, H, Measure, R,
                          Rx, Ry, Rz, S, SqrtX, T, X, Y, Z)
from projectq.ops._basics import BasicGate, BasicRotationGate

from deltalanguage.data_types import DUInt, DSize

from . import IQuantumSimulator
from ..hal import Masks, Opcode, Shifts


class SxGate(BasicGate):
    """Gate that consists of consecutive S and X gate
    (pi-rotation with axis in x-y-plane).
    """

    @property
    def matrix(self):
        return np.array([[0, 1], [1j, 0]])

    def __str__(self):
        return "SX"


#: Shortcut (instance of) :class:`projectq.ops.SxGate`
Sx = SxGate()


class SyGate(BasicGate):
    """Gate that consists of consecutive S and Y gate
    (pi-rotation with axis in x-y-plane).
    """

    @property
    def matrix(self):
        return np.array([[0, 1j], [1, 0]])

    def __str__(self):
        return "SY"


#: Shortcut (instance of) :class:`projectq.ops.SyGate`
Sy = SyGate()


class PiXY(BasicRotationGate):
    """Pi-rotation with axis in x-y-plane gate class."""

    @property
    def matrix(self):
        return np.array([[0, -np.sin(self.angle) - 1j * np.cos(self.angle)],
                         [np.sin(self.angle) - 1j * np.cos(self.angle), 0]])


class PiYZ(BasicRotationGate):
    """Pi-rotation with axis in y-z-plane gate class."""

    @property
    def matrix(self):
        return np.array([[np.cos(self.angle), -1j * np.sin(self.angle)],
                         [1j * np.sin(self.angle), -1 * np.cos(self.angle)]])


class PiZX(BasicRotationGate):
    """Pi-rotation with axis in z-x-plane gate class."""

    @property
    def matrix(self):
        return np.array([[np.cos(self.angle), np.sin(self.angle)],
                         [np.sin(self.angle), -1 * np.cos(self.angle)]])


class ProjectqQuantumSimulator(IQuantumSimulator):
    """Concrete ProjectQ implementation of the IQuantumSimulator interface.

    Parameters
    ----------
    register_size: int
        Size of the qubit register.
    seed : int
        Random number generator seed for both the ProjectQ Simulator and
        circuit errors.
    backend
        ProjectQ backend, could use CircuitDrawer for debugging purposes.


    .. TODO::
        Add and example how to use it in a graph (without any simulator).
    """

    def __init__(self,
                 register_size: int = 16,
                 seed: int = None,
                 backend=Simulator):
        if backend == Simulator and seed is not None:
            self._engine = MainEngine(backend=Simulator(rnd_seed=seed))

        else:
            self._engine = MainEngine(backend=backend())
        self.backend = backend
        self.seed = seed

        # if random numbers are needed to simulate quantum noise use this
        # state in the following way self._random_state.rand()
        self._random_state = RandomState(seed)

        self._qubit_register = None

        # defaulted to 16 because the bitcode status return
        # has 16 bits assigned for measurement results.
        self._qubit_register_size = register_size

        # stores control qubits
        self._control_qubit_indices = []

        # assign projectq gate to each opcode
        self._parameterised_gate_dict = {
            Opcode['CONTROL'].value: C,
            Opcode['R'].value: R,
            Opcode['RX'].value: Rx,
            Opcode['RY'].value: Ry,
            Opcode['RZ'].value: Rz,
            Opcode['PIXY'].value: PiXY,
            Opcode['PIYZ'].value: PiYZ,
            Opcode['PIZX'].value: PiZX,
        }

        self._constant_gate_dict = {
            Opcode['H'].value: H,
            Opcode['S'].value: S,
            Opcode['SQRT_X'].value: SqrtX,
            Opcode['T'].value: T,
            Opcode['X'].value: X,
            Opcode['Y'].value: Y,
            Opcode['Z'].value: Z,
            Opcode['INVS'].value: DaggeredGate(S),
            Opcode['SX'].value: Sx,  # consecutive S and X gate, needed for RC
            Opcode['SY'].value: Sy,  # consecutive S and Y gate, needed for RC
        }
        atexit.register(self.cleanup)

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains
        # all our instance attributes. Always use the dict.copy()
        # method to avoid modifying the original state.
        state = self.__dict__.copy()
        # Remove the engine has it generally unpickable.
        del state['_engine']
        return state

    def __setstate__(self, state):
        # Restore instance attributes (i.e., filename and lineno).
        self.__dict__.update(state)
        # Restore the engine
        if self.backend == Simulator and self.seed is not None:
            self._engine = MainEngine(backend=Simulator(rnd_seed=self.seed))
        else:
            self._engine = MainEngine(backend=self.backend())


    def cleanup(self):
        """Release all the qubits that haven't been handled yet."""
        if self._qubit_register is not None:
            All(Measure) | self._qubit_register
        self._engine.flush()

    def apply_gate(self,
                   gate: BasicGate,
                   qubit_index: int,
                   parameter: float = None):
        """Receives command information and implements the gate on the
        corresponding qubit.

        Parameters
        ----------
        gate : BasicGate
            ProjectQ gate to be applied.
        qubit_index : int
            Index of qubit for gate to be applied to.
        parameter : float
            Angle of gate if parametrised.
        """
        if self._qubit_register is not None:

            if gate is C:
                # add qubit index to self.control_qubit_indices, store
                # in memory until a gate is called, which is run controlled
                # on these qubits.

                if qubit_index in self._control_qubit_indices:
                    raise ValueError(
                        f"Qubit {qubit_index} already set-up as control qubit!"
                    )

                if len(self._control_qubit_indices) + 1 \
                        == self._qubit_register_size:

                    raise ValueError(
                        "Too many control qubits for register size of " +
                        "f{self._qubit_register_size}!"
                    )

                self._control_qubit_indices += [qubit_index]

            else:
                if len(self._control_qubit_indices) == 0:  # single qubit gate
                    if parameter is not None:
                        gate(parameter) | self._qubit_register[qubit_index]
                    else:
                        gate | self._qubit_register[qubit_index]

                else:  # controlled gate
                    if qubit_index in self._control_qubit_indices:
                        raise ValueError(
                            f"Target qubit {qubit_index} already set-up as " +
                            "control qubit!"
                        )

                    control_number = len(self._control_qubit_indices)
                    control_qubits = [
                        self._qubit_register[index] for
                        index in self._control_qubit_indices
                    ]

                    control_reg = tuple(
                        [control_qubits, self._qubit_register[qubit_index]]
                    )
                    if parameter is not None:
                        C(gate(parameter), n=control_number) | control_reg
                    else:
                        C(gate, n=control_number) | control_reg

                    # reset control indices
                    self._control_qubit_indices = []

                self._engine.flush()

    def accept_command(
        self,
        command: DUInt(DSize(32))
    ) -> DUInt(DSize(32)):

        op = command >> Shifts.OPCODE.value
        qubit_index = (command & Masks.QUBIT_INDEX.value)

        if qubit_index + 1 > self._qubit_register_size:
            raise ValueError(
                f"Qubit index ({qubit_index}) greater than " +
                f"register size ({self._qubit_register_size})!"
            )

        if op == Opcode["STATE_PREPARATION"].value:
            if not self._qubit_register:
                self._qubit_register = self._engine.allocate_qureg(
                    self._qubit_register_size
                )

        elif op == Opcode["STATE_MEASURE"].value:

            All(Measure) | self._qubit_register
            self._engine.flush()

            # Each measurement sent should have all valid flags.
            # Therefore valid mask added to the 16bit measurement bitcode.
            measurement_binary = Masks.VALIDS.value

            for n, i in enumerate(self._qubit_register, 0):
                m = int(i)
                measurement_binary += m*2**n

            self._qubit_register = None

            return measurement_binary

        elif op in self._parameterised_gate_dict.keys():
            angle = (command & Masks.ARG.value) >> Shifts.ARG.value
            angle *= (2 * np.pi) / 1024
            gate = self._parameterised_gate_dict[op]

            self.apply_gate(gate, qubit_index, angle)

        elif op in self._constant_gate_dict.keys():
            gate = self._constant_gate_dict[op]
            self.apply_gate(gate, qubit_index)

        elif op == Opcode['ID'].value:
            pass

        else:
            raise TypeError(f"{op} is not a recognised opcode!")



"""Characterization test BaseDeltaType and its subclasses."""

import random
from typing import Any, List, Tuple, Union
import unittest

import attr
import numpy as np

from test._utils import TwoIntsT

from deltalanguage._utils import NamespacedName
from deltalanguage.data_types import (
    BaseDeltaType,
    DeltaTypeError,
    DArray,
    DBool,
    DChar,
    DComplex,
    DFloat,
    DInt,
    DRaw,
    DRecord,
    DSize,
    DStr,
    DTuple,
    DUInt,
    DUnion,
    ForkedReturn,
    Void,
    Top,
    as_delta_type,
    delta_type
)
from deltalanguage.wiring import DeltaGraph


@attr.s(slots=True)
class RecBI:

    x: bool = attr.ib()
    y: int = attr.ib()


@attr.s(slots=True)
class RecBI_copy:

    x: bool = attr.ib()
    y: int = attr.ib()


@attr.s(slots=True)
class RecBDi:

    x: bool = attr.ib()
    y: DInt() = attr.ib()


@attr.s(slots=True)
class RecBT:

    x: bool = attr.ib()
    y: Top() = attr.ib()


@attr.s(slots=True)
class RecIB:

    x: int = attr.ib()
    x: bool = attr.ib()


@attr.s(slots=True)
class RecBIS:

    x: bool = attr.ib()
    y: int = attr.ib()
    z: str = attr.ib()


@attr.s(slots=True)
class RecBII:

    x: bool = attr.ib()
    y: int = attr.ib()
    z: int = attr.ib()


@attr.s(slots=True)
class RecIT:

    x: float = attr.ib()
    y: DTuple([int, int]) = attr.ib()


@attr.s(slots=True)
class RecATI:

    x: DArray(int, DSize(2)) = attr.ib()
    y: DTuple([float, int]) = attr.ib()
    z: int = attr.ib()


class DeltaTypesTest(unittest.TestCase):
    """The most basic tests of BaseDeltaType outside of the context of
    wires (streams, channels) and DeltaGraph.
    """

    def test_DSize(self):
        """Test various use DSize."""
        with self.assertRaises(DeltaTypeError):
            DInt(5)

        with self.assertRaises(ValueError):
            DSize(-1)

        with self.assertRaises(ValueError):
            dummy = DSize(4) + DSize(NamespacedName("a", "b"))

        # add
        d16_32 = DSize(16)
        d16_32 += DSize(32)
        self.assertEqual(d16_32, DSize(48))
        self.assertEqual(DSize(4) + DSize(5), DSize(9))

        # sub
        d32_16 = DSize(32)
        d32_16 -= DSize(16)
        self.assertEqual(d32_16, DSize(16))
        self.assertEqual(DSize(5) - DSize(2), DSize(3))
        with self.assertRaises(ValueError):
            dummy = DSize(5) - DSize(6)

        # mul
        self.assertEqual(DSize(4) * 5, DSize(20))
        self.assertEqual(5 * DSize(4), DSize(20))
        d16x4 = DSize(16)
        d16x4 *= 4
        self.assertEqual(d16x4, DSize(64))
        d16x4 = DSize(16)
        d16x4 *= DSize(4)
        self.assertEqual(d16x4, DSize(64))

        # comparison
        self.assertTrue(DSize(8) == DSize(8))
        self.assertTrue(DSize(8) > DSize(6))
        self.assertTrue(DSize(4) < DSize(6))

    def test_size(self):
        """Test how many bits each data type takes."""
        # primitive
        self.assertEqual(DInt().size, DSize(32))
        self.assertEqual(DUInt().size, DSize(32))
        self.assertEqual(DBool().size, DSize(1))
        self.assertEqual(DChar().size, DSize(8))
        self.assertEqual(DFloat().size, DSize(32))

        # compound
        self.assertEqual(DTuple([int, bool]).size, DSize(33))
        self.assertEqual(DArray(int, DSize(10)).size, DSize(320))
        self.assertEqual(DStr().size, DSize(8192))
        self.assertEqual(DRecord(RecBI).size, DSize(33))

        # compound: DUnion
        self.assertEqual(DUnion([bool]).size, DSize(9))
        self.assertEqual(DUnion([int, bool]).size, DSize(40))
        self.assertEqual(DUnion([int, DTuple([int, int])]).size, DSize(2*32+8))

    def test_str(self):
        """Test string representation of data types."""
        # primitive
        self.assertEqual(str(DInt()), "DInt32")
        self.assertEqual(str(DInt(DSize(64))), "DInt64")
        self.assertEqual(str(DUInt()), "DUInt32")
        self.assertEqual(str(DUInt(DSize(64))), "DUInt64")
        self.assertEqual(str(DBool()), "DBool")
        self.assertEqual(str(DChar()), "DChar8")
        self.assertEqual(str(DFloat()), "DFloat32")
        self.assertEqual(str(DFloat(DSize(64))), "DFloat64")

        # compound
        self.assertEqual(str(DArray(int, DSize(8))), "[DInt32 x 8]")
        self.assertEqual(str(DStr()), "DStr8192")
        self.assertEqual(str(DStr(DSize(100))), "DStr800")
        self.assertEqual(str(DTuple([int, bool])), "(DInt32, DBool)")
        self.assertEqual(str(DRecord(RecBIS)),
                         "{x: DBool, y: DInt32, z: DStr8192}")
        self.assertEqual(str(DUnion([int, bool])), "<DBool | DInt32>")

        # compound: DUnion
        self.assertEqual(str(DUnion([int])), "<DInt32>")
        self.assertEqual(str(DUnion([int, DUnion([int, bool])])),
                         "<DBool | DInt32>")
        self.assertEqual(str(DUnion([int, DUnion([int, DUnion([int, bool])])])),
                         "<DBool | DInt32>")

        # encapsulation of various types
        self.assertEqual(str(DUnion([int, DTuple([int, bool])])),
                         "<(DInt32, DBool) | DInt32>")
        self.assertEqual(str(DArray(DTuple([int, bool]), DSize(8))),
                         "[(DInt32, DBool) x 8]")

        # special
        self.assertEqual(str(Top()), "T")
        self.assertEqual(str(DSize(5)), "5")
        self.assertEqual(str(DSize(NamespacedName("a", "b"))), "(a.b)")
        self.assertEqual(str(ForkedReturn(dict(x=int, y=bool, z=str))),
                         "ForkedReturn(x:DInt32, y:DBool, z:DStr8192)")

    def test_as_python_type(self):
        """Test conversion of Deltaflow data types to python."""
        # special
        self.assertEqual(Top().as_python_type(), Any)

        # primitive
        self.assertEqual(DInt(DSize(32)).as_python_type(), int)
        self.assertEqual(DInt(DSize(64)).as_python_type(), int)
        self.assertEqual(DUInt(DSize(32)).as_python_type(), int)
        self.assertEqual(DUInt(DSize(64)).as_python_type(), int)
        self.assertEqual(DBool().as_python_type(), bool)
        with self.assertRaises(NotImplementedError):
            DChar().as_python_type()
        self.assertEqual(DFloat(DSize(32)).as_python_type(), float)
        self.assertEqual(DFloat(DSize(64)).as_python_type(), float)
        self.assertEqual(DComplex(DSize(64)).as_python_type(), complex)
        self.assertEqual(DComplex(DSize(128)).as_python_type(), complex)

        # compound
        self.assertEqual(DTuple([int, bool]).as_python_type(),
                         Tuple[int, bool])
        self.assertEqual(DTuple([int, DTuple([int, bool])]).as_python_type(),
                         Tuple[int, Tuple[int, bool]])
        self.assertEqual(DArray(int, DSize(3)).as_python_type(),
                         List[int])

        self.assertEqual(DStr().as_python_type(), str)
        self.assertEqual(DStr(DSize(10)).as_python_type(), str)

        self.assertEqual(DRecord(RecBI).as_python_type(), RecBI)
        self.assertEqual(DRecord(RecBDi).as_python_type(), RecBDi)
        self.assertNotEqual(DRecord(RecBI).as_python_type(), RecBI_copy)

        # compound: DUnion
        self.assertEqual(DUnion([bool, int]).as_python_type(),
                         Union[bool, int])
        self.assertEqual(DUnion([bool, DTuple([int, bool])]).as_python_type(),
                         Union[bool, Tuple[int, bool]])

    def test_types_comparison(self):
        """Various tests of types comparison."""
        # primitive
        self.assertEqual(DInt(DSize(32)), DInt())
        self.assertNotEqual(DInt(), DUInt())
        self.assertNotEqual(DInt(), DInt(DSize(64)))

        # compound
        self.assertEqual(DTuple([int, bool]), DTuple([int, bool]))
        self.assertNotEqual(DTuple([int, bool]), DTuple([bool, int]))
        self.assertEqual(DArray(int, DSize(4)), DArray(int, DSize(4)))
        self.assertEqual(DArray(int, DSize(4)), DArray(DInt(), DSize(4)))
        self.assertNotEqual(DArray(int, DSize(4)), DArray(int, DSize(5)))
        self.assertNotEqual(DStr(), DStr(DSize(100)))
        self.assertEqual(DRecord(RecBI), DRecord(RecBI))

        # compound: DUnion
        self.assertEqual(DUnion([int, bool]), DUnion([bool, int]))
        self.assertEqual(DUnion([int, DUnion([int, bool])]),
                         DUnion([int, bool]))
        self.assertEqual(DUnion([int, DUnion([int, DUnion([int, bool])])]),
                         DUnion([int, bool]))
        self.assertEqual(DUnion([int, int]), DUnion([int]))
        self.assertNotEqual(DUnion([DInt()]), DInt())

        # special
        self.assertEqual(ForkedReturn(dict(x=int, y=bool, z=str)),
                         ForkedReturn(dict(x=int, y=bool, z=str)))

    def test_DUnion(self):
        """DUnion specific types."""

        with self.assertRaises(DeltaTypeError):
            DUnion([])

        with self.assertRaises(DeltaTypeError):
            DUnion([None])

        with self.assertRaises(DeltaTypeError):
            DUnion([None, int])


class DeltaTypesPackTest(unittest.TestCase):
    """Test pack/unpack methods for `BaseDeltaType`."""

    def pack_unpack(self, val, t1, t2=None):
        """Helper that pack and unpack a value via given types."""
        if t2 is None:
            t2 = t1
        buf = t1.pack(val)
        val_new = t2.unpack(buf)
        return val_new

    def check(self, val, t1, t2=None):
        """Test that pack-unpack returns exactly the same message."""
        val_new = self.pack_unpack(val, t1, t2)
        self.assertEqual(val, val_new)

    def check_numpy(self, val, t1):
        val_new = self.pack_unpack(t1.as_numpy_object(val), t1)
        self.assertEqual(val, val_new)

    def check_float(self, val, t: DFloat):
        """Test pack-unpack for floats."""
        if not isinstance(t, DFloat):
            raise DeltaTypeError

        # TODO this check can be done on binary, then the number of places
        # will be more reasonably explained
        if t.size == DSize(32):
            places = 7
        elif t.size == DSize(64):
            places = 15
        else:
            raise NotImplementedError('Unsupported format')

        val_new = self.pack_unpack(val, t)
        self.assertAlmostEqual(val, val_new, places=places)

    def check_complex(self, val, t: DComplex):
        """Test pack-unpack for complex numbers."""
        if not isinstance(t, DComplex):
            raise DeltaTypeError

        # Using the same idea as check_float, but for real and imaginary parts
        if t.size == DSize(64):
            places = 7
        elif t.size == DSize(128):
            places = 15
        else:
            raise NotImplementedError('Unsupported format')

        val_new = self.pack_unpack(val, t)
        self.assertAlmostEqual(val.real, val_new.real, places=places)
        self.assertAlmostEqual(val.imag, val_new.imag, places=places)

    def test_DInt(self):
        """Only 8, 16, 32, and 64 bits are supported."""
        for bits in (8, 16, 32, 64):
            self.check(-2**(bits-1), DInt(DSize(bits)))
            self.check(2**(bits-1) - 1, DInt(DSize(bits)))
            for _ in range(1000):
                self.check(random.randint(-2**(bits-1), 2**(bits-1) - 1),
                           DInt(DSize(bits)))

    def test_DUInt(self):
        """Only 8, 16, 32, and 64 bits are supported."""
        for bits in (8, 16, 32, 64):
            self.check(0, DUInt(DSize(bits)))
            self.check(2**bits - 1, DUInt(DSize(bits)))
            for _ in range(1000):
                self.check(random.randint(0, 2**bits - 1),
                           DUInt(DSize(bits)))

    def test_DBool(self):
        self.check(False, DBool())
        self.check(0, DBool())
        self.check(True, DBool())
        self.check(1, DBool())

    def test_DFloat(self):
        """Only 32 and 64 bits are supported."""
        precision_dict = {32: -23, 64: -52}

        for bits, precision in precision_dict.items():
            for _ in range(1000):
                self.check_float(random.uniform(-1, 1), DFloat(DSize(bits)))

            self.check(1 + 2**precision, DFloat(DSize(bits)))

    def test_DComplex(self):
        """Only 64 and 128 bits are supported."""
        for bits in (64, 128):
            for _ in range(1000):
                self.check_complex(random.uniform(-1, 1) +
                                   random.uniform(-1, 1) * 1j,
                                   DComplex(DSize(bits)))

    def test_DArray(self):
        # primitive elements are properly handled
        # int are passed as DInt, not DUInt
        self.check([1, 2, 3], DArray(DInt(), DSize(3)))
        with self.assertRaises(DeltaTypeError):
            self.check([1, 2, 3], DArray(DUInt(), DSize(3)))

        # for floats use a dot
        # might be a potential problem, due to python silent type downcasting
        self.check([1.0, 2.0, 3.0], DArray(DFloat(), DSize(3)))
        with self.assertRaises(DeltaTypeError):
            self.check([1, 2, 3], DArray(DFloat(), DSize(3)))

        # bool are passed as DBool, not DInt
        self.check([True, False, False], DArray(DBool(), DSize(3)))
        with self.assertRaises(DeltaTypeError):
            self.check([1, 0, 0], DArray(DBool(), DSize(3)))

        # encapsulation of compound types
        self.check([[1, 2, 3], [4, 5, 6]],
                   DArray(DArray(DInt(), DSize(3)), DSize(2)))

        with self.assertRaises(DeltaTypeError):
            self.check([1, 2, 3, 4, 5, 6],
                       DArray(DArray(DInt(), DSize(3)), DSize(2)))

        with self.assertRaises(AssertionError):
            self.check([1, 2, 3, 4, 5, 6],
                       DArray(DInt(), DSize(6)),
                       DArray(DArray(DInt(), DSize(3)), DSize(2)))

        # mixed types
        self.check([(1, 2, 3), (4, 5, 6)],
                   DArray(DTuple([int, int, int]), DSize(2)))

        self.check(["hello", "world"], DArray(DStr(DSize(5)), DSize(2)))

        # numpy
        self.check_numpy([1, 2, 3, 4, 5], DArray(int, DSize(5)))

    def test_DStr(self):
        self.check('hello world', DStr())
        self.check('A' * 1024, DStr())
        self.check('check digits 14213', DStr())
        self.check('check spaces in the end ', DStr())

        with self.assertRaises(DeltaTypeError):
            self.check('123456', DStr(DSize(4)))

        self.check((-5, 'text'), DTuple([int, DStr()]))
        self.check(['hello', 'world!'], DArray(DStr(), DSize(2)))

        self.check_numpy('hello world', DStr())

    def test_DTuple(self):
        # primitive elements are properly handled
        self.check((-5, True, 3.25), DTuple([int, bool, float]))

        with self.assertRaises(DeltaTypeError):
            self.check((-5, True, 3.25), DTuple([int, bool, int]))

        # incapsulation
        self.check((-5, (1, 2)), DTuple([int, DTuple([int, int])]))

        with self.assertRaises(AssertionError):
            self.check((-5, (1, 2)),
                       DTuple([int, DTuple([int, int])]),
                       DTuple([int, int, int]))

        # mixed types
        self.check(([1, 2, 3], [4.0, 5.0]),
                   DTuple([DArray(int, DSize(3)), DArray(float, DSize(2))]))

        self.check(("hello", "world"), DTuple([DStr(), DStr(DSize(6))]))

        # numpy
        self.check_numpy((1, 2.0, True), DTuple([int, float, bool]))

    def test_DRecord(self):
        # primitive
        self.check(RecBI(True, 5), DRecord(RecBI))
        self.check(-4, DInt())
        self.check(RecBII(True, 5, -4), DRecord(RecBII))
        with self.assertRaises(DeltaTypeError):
            self.check(RecBI(True, 5), DRecord(RecIB))

        # mixed
        self.check(RecIT(-4.0, (1, 2)), DRecord(RecIT))
        self.check(RecATI([1, 2], (3.0, 4), 5),
                   DRecord(RecATI))

        self.check((RecIT(-4.0, (1, 2)), 1),
                   DTuple([DRecord(RecIT), int]))

        self.check([RecIT(-4.0, (1, 2)), RecIT(5.0, (-3, -4))],
                   DArray(DRecord(RecIT), DSize(2)))

        # numpy
        self.check_numpy(RecBI(False, 2), DRecord(RecBI))

    def test_DUnion(self):
        # primitive
        self.check(5, DUnion([int, bool]), DUnion([int, bool]))
        self.check(True, DUnion([int, bool]), DUnion([bool, int]))

        # compound
        self.check(5, DUnion([int, DTuple([int, float])]))
        self.check((4, 5), DUnion([int, DTuple([int, int])]))
        self.check((4, 5),
                   DUnion([DArray(int, DSize(2)), DTuple([int, int])]))
        self.check([4, 5],
                   DUnion([DArray(int, DSize(2)), DTuple([int, int])]))

        # buffer's size is always the same
        self.assertEqual(len(DUnion([int, bool]).pack(5)),
                         DUnion([int, bool]).size.val)
        self.assertEqual(len(DUnion([int, bool]).pack(True)),
                         DUnion([int, bool]).size.val)

        # numpy (throws error)
        with self.assertRaises(
                DeltaTypeError,
                msg="NumPy unions cannot be converted to Python types."):
            self.check_numpy(5, DUnion([bool, float, int]))


class WiresTest(unittest.TestCase):
    """Testing the rules of data transmission in a single wire of DeltaGraph."""

    def test_primitive_types(self):
        """Strict typing without subtyping."""
        self.assertTrue(DeltaGraph.check_wire(DInt(), DInt()))

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DInt(), DUInt())

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DUInt(), DInt())

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(int, int)

    def test_Top(self):
        """Everything can be accepted as Top()."""
        self.assertTrue(DeltaGraph.check_wire(DInt(), Top()))
        self.assertTrue(DeltaGraph.check_wire(DUInt(), Top()))
        self.assertTrue(DeltaGraph.check_wire(DBool(), Top()))
        self.assertTrue(DeltaGraph.check_wire(DTuple([int, bool]), Top()))
        self.assertTrue(DeltaGraph.check_wire(DUnion([int, bool]), Top()))
        self.assertTrue(DeltaGraph.check_wire(DArray(int, DSize(8)), Top()))
        self.assertTrue(DeltaGraph.check_wire(DStr(), Top()))
        self.assertTrue(DeltaGraph.check_wire(DRecord(RecBI), Top()))
        self.assertTrue(DeltaGraph.check_wire(Top(), Top()))

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(Top(), DInt())

        # however it's not true if Top is used within a non-primitive type
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DTuple([int, int]), DTuple([int, Top()]))
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DArray(int, DSize(8)),
                                  DArray(Top(), DSize(8)))
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DRecord(RecBI), DRecord(RecBT))

    def test_DUnion(self):
        """Test wires with DUnion."""
        # examples of obvious behaiviour
        self.assertTrue(DeltaGraph.check_wire(DUnion([int, bool]),
                                              DUnion([int, bool])))
        self.assertTrue(DeltaGraph.check_wire(DUnion([int, bool]),
                                              DUnion([bool, int])))

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DUnion([int, bool]), DInt())
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DUnion([int, bool, float]),
                                  DUnion([int, bool]))

        # strict typing even with DUnion, i.e. all subtypes should match
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DUnion([bool, int]),
                                  DUnion([bool, int, float]))

        # DUnion changes packing method, thus these tests should fail
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DInt(), DUnion([int]))
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DUnion([int]), DInt())

    def test_DArray(self):
        """Only strict typing."""
        self.assertTrue(DeltaGraph.check_wire(DArray(int, DSize(8)),
                                              DArray(int, DSize(8))))

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DArray(int, DSize(10)),
                                  DArray(int, DSize(8)))

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DArray(int, DSize(8)),
                                  DArray(int, DSize(10)))

    def test_DTuple(self):
        """Only strict typing."""
        self.assertTrue(DeltaGraph.check_wire(DTuple([int, bool]),
                                              DTuple([int, bool])))

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DTuple([int, bool]),
                                  DTuple([bool, int]))

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DTuple([int, bool]),
                                  DTuple([int, bool, bool]))

    def test_DRecord(self):
        """Only strict typing."""
        self.assertTrue(DeltaGraph.check_wire(DRecord(RecBI), DRecord(RecBI)))

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DRecord(RecBI), DRecord(RecIB))

        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DRecord(RecBIS), DRecord(RecBI))

    def test_ForkedReturn(self):
        self.assertTrue(DeltaGraph.check_wire(TwoIntsT.elem_dict['x'],
                                              DInt()))


class UtilsTest(unittest.TestCase):
    """Helper functions."""

    def test_as_delta_type(self):
        """Test conversion from python to Deltaflow data types."""
        # special
        self.assertEqual(as_delta_type(object), Top())
        self.assertEqual(as_delta_type(type(object)), Top())
        self.assertEqual(as_delta_type(type), Top())

        self.assertEqual(as_delta_type(Void), Void)

        with self.assertRaises(DeltaTypeError):
            as_delta_type(None)
        with self.assertRaises(DeltaTypeError):
            as_delta_type(type(None))

        # primitive
        self.assertNotEqual(as_delta_type(bool), DUInt(DSize(1)))
        self.assertEqual(as_delta_type(bool), DBool())
        self.assertEqual(as_delta_type(np.bool_), DBool())
        self.assertEqual(as_delta_type(int), DInt(DSize(32)))
        self.assertEqual(as_delta_type(np.int8), DChar())
        self.assertEqual(as_delta_type(np.int16), DInt(DSize(16)))
        self.assertEqual(as_delta_type(np.int32), DInt(DSize(32)))
        self.assertEqual(as_delta_type(np.int64), DInt(DSize(64)))
        self.assertEqual(as_delta_type(np.uint8), DChar())
        self.assertEqual(as_delta_type(np.uint16), DUInt(DSize(16)))
        self.assertEqual(as_delta_type(np.uint32), DUInt(DSize(32)))
        self.assertEqual(as_delta_type(np.uint64), DUInt(DSize(64)))
        self.assertEqual(as_delta_type(float), DFloat())
        self.assertEqual(as_delta_type(np.float32), DFloat(DSize(32)))
        self.assertEqual(as_delta_type(np.float64), DFloat(DSize(64)))
        self.assertEqual(as_delta_type(complex), DComplex())
        self.assertEqual(as_delta_type(np.complex64), DComplex(DSize(64)))
        self.assertEqual(as_delta_type(np.complex128), DComplex(DSize(128)))

        # compound
        with self.assertRaises(DeltaTypeError):
            as_delta_type(Tuple[int, bool])
        with self.assertRaises(DeltaTypeError):
            as_delta_type(List[int])
        self.assertNotEqual(as_delta_type(str), DArray(DChar(), DSize(1024)))
        self.assertEqual(as_delta_type(str), DStr())
        self.assertEqual(as_delta_type(RecBI), DRecord(RecBI))

        # numpy compound
        self.assertEqual(as_delta_type(DArray(int, DSize(5)).as_numpy_type()),
                         DArray(int, DSize(5)))
        self.assertEqual(as_delta_type(DStr().as_numpy_type()), DStr())
        self.assertEqual(
            as_delta_type(DTuple([int, bool, float]).as_numpy_type()),
            DTuple([int, bool, float])
        )
        self.assertEqual(as_delta_type(DRecord(RecBI).as_numpy_type()),
                         DRecord(RecBI))
        self.assertEqual(
            as_delta_type(DUnion([bool, float, int]).as_numpy_type()),
            DUnion([bool, float, int]))

    def test_delta_type(self):
        """Test mapping python objects to Deltaflow data types."""
        # special
        with self.assertRaises(DeltaTypeError):
            delta_type(None)

        # primitive
        self.assertEqual(delta_type(False), DBool())
        self.assertEqual(delta_type(np.bool_(False)), DBool())
        self.assertEqual(delta_type(5), DInt(DSize(32)))
        self.assertEqual(delta_type(np.int16(5)), DInt(DSize(16)))
        self.assertEqual(delta_type(np.int32(5)), DInt(DSize(32)))
        self.assertEqual(delta_type(np.int64(5)), DInt(DSize(64)))
        self.assertEqual(delta_type(np.uint16(5)), DUInt(DSize(16)))
        self.assertEqual(delta_type(np.uint32(5)), DUInt(DSize(32)))
        self.assertEqual(delta_type(np.uint64(5)), DUInt(DSize(64)))
        self.assertEqual(delta_type(4.2), DFloat(DSize(32)))
        self.assertEqual(delta_type(np.float32(4.2)), DFloat(DSize(32)))
        self.assertEqual(delta_type(np.float64(4.2)), DFloat(DSize(64)))
        self.assertEqual(delta_type(3+1j), DComplex(DSize(64)))
        self.assertEqual(delta_type(np.complex64(3+1j)), DComplex(DSize(64)))
        self.assertEqual(delta_type(np.complex128(3+1j)), DComplex(DSize(128)))
        self.assertEqual(delta_type('c'), DChar())

        # compound
        self.assertEqual(delta_type((1, True, 3.7)),
                         DTuple([int, bool, float]))
        self.assertEqual(delta_type([1, 2, 4]), DArray(int, DSize(3)))
        self.assertEqual(delta_type(RecBI(True, 5)), DRecord(RecBI))

        # numpy compound
        self.assertEqual(delta_type(np.array([1, 2, 3, 4, 5])),
                         DArray(DInt(DSize(64)), DSize(5)))
        self.assertEqual(delta_type(np.array([1, 2.0, 3, 4, 5])),
                         DArray(DFloat(DSize(64)), DSize(5)))
        self.assertEqual(delta_type(
            DStr(DSize(5)).as_numpy_object("abcde")), DStr(DSize(5)))
        self.assertEqual(
            delta_type(DTuple([int, float, bool]
                              ).as_numpy_object((1, 2.0, True))),
            DTuple([int, float, bool])
        )
        self.assertEqual(
            delta_type(DRecord(RecBI).as_numpy_object(RecBI(True, 2))),
            DRecord(RecBI)
        )
        self.assertEqual(
            delta_type(DUnion([bool, float, int]).as_numpy_object(5.0)),
            DUnion([bool, float, int])
        )

        # different combinations
        self.assertEqual(delta_type([(4, 4.3), (2, 3.3)]),
                         DArray(DTuple([int, float]), DSize(2)))


class DeltaTypesNumpyTest(unittest.TestCase):
    """Test from_numpy_object/as_numpy_object methods for `BaseDeltaType`.

    .. todo::
        Can be rewritten to inherit from DeltaTypesPackTest.
        The only possible exception is the DUnion test as that one is only
        one-way.
    """

    def to_np_and_back(self, val, t):
        return t.from_numpy_object(t.as_numpy_object(val))

    def check(self, val, t):
        self.assertEqual(self.to_np_and_back(val, t), val)

    def check_float(self, val, t: DFloat):
        """Test pack-unpack for floats."""
        if not isinstance(t, DFloat):
            raise DeltaTypeError

        # TODO this check can be done on binary, then the number of places
        # will be more reasonably explained
        if t.size == DSize(32):
            places = 6
        elif t.size == DSize(64):
            places = 15
        else:
            raise NotImplementedError('Unsupported format')

        val_new = self.to_np_and_back(val, t)
        self.assertAlmostEqual(val, val_new, places=places)

    def check_complex(self, val, t: DComplex):
        """Test pack-unpack for complex numbers."""
        if not isinstance(t, DComplex):
            raise DeltaTypeError

        # Using the same idea as check_float, but for real and imaginary parts
        if t.size == DSize(64):
            places = 6
        elif t.size == DSize(128):
            places = 15
        else:
            raise NotImplementedError('Unsupported format')

        val_new = self.to_np_and_back(val, t)
        self.assertAlmostEqual(val.real, val_new.real, places=places)
        self.assertAlmostEqual(val.imag, val_new.imag, places=places)

    def test_Top(self):
        with self.assertRaises(NotImplementedError):
            Top().as_numpy_type()

        with self.assertRaises(NotImplementedError):
            Top().as_numpy_object(True)

        with self.assertRaises(NotImplementedError):
            Top().from_numpy_object(np.bool_(False))

    def test_DBool(self):
        self.assertEqual(DBool().as_numpy_type(), np.bool_)

    def test_DBool_object(self):
        self.check(False, DBool())

    def test_DInt_type(self):
        self.assertEqual(DInt(DSize(8)).as_numpy_type(), np.int8)
        self.assertEqual(DInt(DSize(16)).as_numpy_type(), np.int16)
        self.assertEqual(DInt(DSize(32)).as_numpy_type(), np.int32)
        self.assertEqual(DInt(DSize(64)).as_numpy_type(), np.int64)
        with self.assertRaises(NotImplementedError):
            DInt(DSize(128)).as_numpy_type()

    def test_DInt_object(self):
        self.check(5, DInt(DSize(8)))
        self.check(10, DInt(DSize(16)))
        self.check(15, DInt(DSize(32)))
        self.check(20, DInt(DSize(64)))

    def test_DUInt_type(self):
        self.assertEqual(DUInt(DSize(8)).as_numpy_type(), np.uint8)
        self.assertEqual(DUInt(DSize(16)).as_numpy_type(), np.uint16)
        self.assertEqual(DUInt(DSize(32)).as_numpy_type(), np.uint32)
        self.assertEqual(DUInt(DSize(64)).as_numpy_type(), np.uint64)
        with self.assertRaises(NotImplementedError):
            DUInt(DSize(128)).as_numpy_type()

    def test_DUInt_object(self):
        self.check(5, DUInt(DSize(8)))
        self.check(10, DUInt(DSize(16)))
        self.check(15, DUInt(DSize(32)))
        self.check(20, DUInt(DSize(64)))

    def test_DChar_type(self):
        self.assertEqual(DChar().as_numpy_type(), np.uint8)

    def test_DChar_object(self):
        self.check('a', DChar())

    def test_DFloat_type(self):
        self.assertEqual(DFloat(DSize(32)).as_numpy_type(), np.float32)
        self.assertEqual(DFloat(DSize(64)).as_numpy_type(), np.float64)
        with self.assertRaises(NotImplementedError):
            DUInt(DSize(128)).as_numpy_type()

    def test_DFloat_object(self):
        self.check_float(3.14, DFloat(DSize(32)))
        self.check_float(6.28, DFloat(DSize(64)))

    def test_DComplex_type(self):
        self.assertEqual(DComplex(DSize(64)).as_numpy_type(), np.complex64)
        self.assertEqual(DComplex(DSize(128)).as_numpy_type(), np.complex128)

    def test_DComplex_object(self):
        self.check_complex(1 + 0.5j, DComplex(DSize(64)))
        self.check_complex(5 + 0.2j, DComplex(DSize(128)))

    def test_DArray_type(self):
        int_type = DArray(int, DSize(5)).as_numpy_type()
        self.assertEqual(int_type[0], (np.int32, (5,)))

    def test_DArray_object(self):
        self.check([1, 2, 3, 4, 5], DArray(int, DSize(5)))

    def test_DStr_type(self):
        str_type = DStr(DSize(32)).as_numpy_type()
        self.assertEqual(str_type, (np.string_, 32))

    def test_DStr_object(self):
        self.check("abcde", DStr(DSize(32)))

    def test_DTuple_type(self):
        tuple_type = DTuple((int, bool, DChar())).as_numpy_type()
        self.assertEqual(tuple_type[0], np.int32)
        self.assertEqual(tuple_type[1], np.bool_)
        self.assertEqual(tuple_type[2], np.uint8)

    def test_DTuple_object(self):
        t = DTuple((int, bool, DChar()))
        self.check((5, True, 'c'), t)

    def test_DRecord_type(self):
        record_type = DRecord(RecBI).as_numpy_type()
        self.assertEqual(record_type['x'], np.bool_)
        self.assertEqual(record_type['y'], np.int32)

    def test_DRecord_object(self):
        rectangle = RecBI(True, 5)
        self.check(rectangle, DRecord(RecBI))

    def test_DUnion_type(self):
        union_type = DUnion([bool, DChar(), int]).as_numpy_type()
        self.assertEqual(union_type[0], np.bool_)
        self.assertEqual(union_type[1], np.uint8)
        self.assertEqual(union_type[2], np.int32)
        self.assertEqual(union_type.fields['DBool'][1], 0)
        self.assertEqual(union_type.fields['DChar8'][1], 0)
        self.assertEqual(union_type.fields['DInt32'][1], 0)

    def test_DUnion_object(self):
        union_object = DUnion([bool, DChar(), int]).as_numpy_object('c')
        self.assertEqual(union_object[0]["DChar8"], 99)

        with self.assertRaises(
                DeltaTypeError,
                msg="NumPy unions cannot be converted to Python types."):
            self.check('c', DUnion([bool, DChar(), int]))

    def test_compound_objects(self):
        t = DArray(DTuple([bool, int]), DSize(3))
        val = [(True, 1), (False, 2), (True, 3), (False, 4), (True, 5)]
        self.check(val, t)

        t = DTuple([int, DTuple([bool, int])])
        val = (12, (True, 8))
        self.check(val, t)

        t = DTuple([int, DArray(int, DSize(2))])
        val = (12, [14, 18])
        self.check(val, t)

        t = DTuple([int, DStr()])
        val = (12, "hello")
        self.check(val, t)

        t = DTuple([int, DRecord(RecBI)])
        val = (12, RecBI(True, 8))
        self.check(val, t)

        t = DTuple([int, DUnion([bool, int])])
        val = (12, True)
        np_val = t.as_numpy_object(val)
        self.assertEqual(DInt().from_numpy_object(np_val[0][0]), 12)
        self.assertEqual(DBool().from_numpy_object(np_val[0][1][1]), True)

        t = DRecord(RecATI)
        val = RecATI([1, 2], (3.0, 4), 5)
        self.check(val, t)

        t = DUnion([DArray(int, DSize(2)), int])
        val = [1, 2]
        np_val = t.as_numpy_object(val)
        new_val = DArray(int, DSize(2)).from_numpy_object(np_val[0][1])
        self.assertEqual(val, new_val)

        t = DUnion([str, int])
        val = "abcde"
        np_val = t.as_numpy_object(val)
        new_val = DStr().from_numpy_object(np_val[0][1])
        self.assertEqual(val, new_val)


class DRawTest(unittest.TestCase):
    """Test DRaw packing/unpacking + transmission rules.

    .. todo::
        This test case repeats everything from DeltaTypesPackTest but with 
        different check method, specifically for DRaw. This code repetition
        can be reduced via inheritance.

        Plus there are checks how DeltaGraph.check_wire works with DRaw.
        They can be moved to WiresTest.
    """

    def to_and_from_bits(self, val, base_type):
        """Helper that converts a value to and from bits via given type."""
        raw_type = DRaw(base_type)
        return raw_type.from_bits(raw_type.as_bits(val))

    def pack_unpack(self, val, base_type):
        """Packs and unpacks a DRaw type."""
        raw_type = DRaw(base_type)
        packed = raw_type.pack(raw_type.as_bits(val))
        return raw_type.from_bits(raw_type.unpack(packed))

    def check_error(self, val, base_type):
        if not isinstance(val, int):
            with self.assertRaises(DeltaTypeError):
                DRaw(base_type).pack(val)

    def check(self, val, t):
        """Test that to and from bits returns exactly the same message."""
        val_new = self.to_and_from_bits(val, t)
        self.assertEqual(val, val_new)
        val_new = self.pack_unpack(val, t)
        self.assertEqual(val, val_new)
        self.check_error(val, t)

    def check_float(self, val, t: DFloat):
        """Test for floats."""
        if not isinstance(t, DFloat):
            raise DeltaTypeError

        # TODO this check can be done on binary, then the number of places
        # will be more reasonably explained
        if t.size == DSize(32):
            places = 7
        elif t.size == DSize(64):
            places = 15
        else:
            raise NotImplementedError('Unsupported format')

        val_new = self.to_and_from_bits(val, t)
        self.assertAlmostEqual(val, val_new, places=places)
        val_new = self.pack_unpack(val, t)
        self.assertAlmostEqual(val, val_new, places=places)
        self.check_error(val, t)

    def check_complex(self, val, t: DComplex):
        """Test for complex numbers."""
        if not isinstance(t, DComplex):
            raise DeltaTypeError

        # Using the same idea as check_float, but for real and imaginary parts
        if t.size == DSize(64):
            places = 7
        elif t.size == DSize(128):
            places = 15
        else:
            raise NotImplementedError('Unsupported format')

        val_new = self.to_and_from_bits(val, t)
        self.assertAlmostEqual(val.real, val_new.real, places=places)
        self.assertAlmostEqual(val.imag, val_new.imag, places=places)
        val_new = self.pack_unpack(val, t)
        self.assertAlmostEqual(val.real, val_new.real, places=places)
        self.assertAlmostEqual(val.imag, val_new.imag, places=places)
        self.check_error(val, t)

    def test_DInt(self):
        """Only 8, 16, 32 and 64 bits are supported."""
        for bits in (8, 16, 32, 64):
            self.check(-2**(bits-1), DInt(DSize(bits)))
            self.check(2**(bits-1) - 1, DInt(DSize(bits)))
            for _ in range(1000):
                self.check(random.randint(-2**(bits-1), 2**(bits-1) - 1),
                           DInt(DSize(bits)))

        self.assertTrue(DeltaGraph.check_wire(DRaw(int), DRaw(int)))
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DRaw(DInt(DSize(32))),
                                  DRaw(DInt(DSize(64))))

    def test_DUInt(self):
        """Only 8, 16, 32 and 64 bits are supported."""
        for bits in (8, 16, 32, 64):
            self.check(0, DUInt(DSize(bits)))
            self.check(2**bits - 1, DUInt(DSize(bits)))
            for _ in range(1000):
                self.check(random.randint(0, 2**bits - 1),
                           DUInt(DSize(bits)))

        self.assertTrue(DeltaGraph.check_wire(DRaw(DUInt(DSize(32))),
                                              DRaw(DUInt(DSize(32)))))
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DRaw(DUInt(DSize(32))),
                                  DRaw(DUInt(DSize(64))))

    def test_DBool(self):
        self.check(False, DBool())
        self.check(0, DBool())
        self.check(True, DBool())
        self.check(1, DBool())
        self.assertTrue(DeltaGraph.check_wire(DRaw(DBool()), DRaw(DBool())))

    def test_DFloat(self):
        """Only 32 and 64 bits are supported."""
        precision_dict = {32: -23, 64: -52}

        for bits, precision in precision_dict.items():
            for _ in range(1000):
                self.check_float(random.uniform(-1, 1), DFloat(DSize(bits)))

            self.check(1 + 2**precision, DFloat(DSize(bits)))

        self.assertTrue(DeltaGraph.check_wire(DRaw(float), DRaw(float)))
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DRaw(DFloat(DSize(32))),
                                  DRaw(DFloat(DSize(64))))

    def test_DComplex(self):
        """Only 64 and 128 bits are supported."""
        for bits in (64, 128):
            for _ in range(1000):
                self.check_complex(random.uniform(-1, 1) +
                                   random.uniform(-1, 1) * 1j,
                                   DComplex(DSize(bits)))

        self.assertTrue(DeltaGraph.check_wire(DRaw(complex), DRaw(complex)))
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DRaw(DComplex(DSize(64))),
                                  DRaw(DComplex(DSize(128))))

    def test_DArray(self):
        # primitive elements are properly handled
        # int are passed as DInt, not DUInt
        self.check([1, 2, 3], DArray(DInt(), DSize(3)))

        # for floats use a dot
        # might be a potential problem, due to python silent type downcasting
        self.check([1.0, 2.0, 3.0], DArray(DFloat(), DSize(3)))

        # bool are passed as DBool, not DInt
        self.check([True, False, False], DArray(DBool(), DSize(3)))

        # encapsulation of compound types
        self.check([[1, 2, 3], [4, 5, 6]],
                   DArray(DArray(DInt(), DSize(3)), DSize(2)))

        # mixed types
        self.check([(1, 2, 3), (4, 5, 6)],
                   DArray(DTuple([int, int, int]), DSize(2)))

        self.check(["hello", "world"], DArray(DStr(DSize(5)), DSize(2)))
        self.assertTrue(DeltaGraph.check_wire(DRaw(DArray(int, DSize(2))),
                                              DRaw(DArray(int, DSize(2)))))
        with self.assertRaises(DeltaTypeError):
            DeltaGraph.check_wire(DRaw(DArray(int, DSize(2))),
                                  DRaw(DArray(int, DSize(3))))

    def test_DStr(self):
        self.check('hello world', DStr())
        self.check('A' * 1024, DStr())
        self.check('check digits 14213', DStr())
        self.check('check spaces in the end ', DStr())

        self.check((-5, 'text'), DTuple([int, DStr()]))
        self.check(['hello', 'world!'], DArray(DStr(), DSize(2)))
        self.assertTrue(DeltaGraph.check_wire(DRaw(DStr()), DRaw(DStr())))

    def test_DTuple(self):
        # primitive elements are poperly handled
        self.check((-5, True, 3.25), DTuple([int, bool, float]))

        # incapsulation
        self.check((-5, (1, 2)), DTuple([int, DTuple([int, int])]))

        # mixed types
        self.check(([1, 2, 3], [4.0, 5.0]),
                   DTuple([DArray(int, DSize(3)), DArray(float, DSize(2))]))

        self.check(("hello", "world"), DTuple([DStr(), DStr(DSize(6))]))
        self.assertTrue(DeltaGraph.check_wire(DRaw(DTuple([DStr(), int])),
                                              DRaw(DTuple([DStr(), int]))))

    def test_DRecord(self):
        # primitive
        self.check(RecBI(True, 5), DRecord(RecBI))
        self.check(-4, DInt())
        self.check(RecBII(True, 5, -4), DRecord(RecBII))

        # mixed
        self.check(RecIT(-4.0, (1, 2)), DRecord(RecIT))
        self.check(RecATI([1, 2], (3.0, 4), 5),
                   DRecord(RecATI))

        self.check((RecIT(-4.0, (1, 2)), 1),
                   DTuple([DRecord(RecIT), int]))

        self.check([RecIT(-4.0, (1, 2)), RecIT(5.0, (-3, -4))],
                   DArray(DRecord(RecIT), DSize(2)))
        self.assertTrue(DeltaGraph.check_wire(DRaw(DRecord(RecIT)),
                                              DRaw(DRecord(RecIT))))

    def test_DUnion(self):
        # primitive
        self.check(5, DUnion([int, bool]))
        self.check(True, DUnion([int, bool]))

        # compound
        self.check(5, DUnion([int, DTuple([int, float])]))
        self.check((4, 5), DUnion([int, DTuple([int, int])]))
        self.check((4, 5),
                   DUnion([DArray(int, DSize(2)), DTuple([int, int])]))
        self.check([4, 5],
                   DUnion([DArray(int, DSize(2)), DTuple([int, int])]))
        self.assertTrue(DeltaGraph.check_wire(DRaw(DUnion([DStr(), int])),
                                              DRaw(DUnion([DStr(), int]))))


if __name__ == "__main__":
    unittest.main()



class Latency:
    """Latency of a body, how long we expect it to take.

    Parameters
    ----------
    clocks : int
        The expected latency of the node in clock cycles.
    time : int
        The expected latency in real time.
    variance : int
        Variance of the node's latency.

    ..
        This is not used currently. Use cases:

        - latencies can be used to optimise the emulation process
        - latencies can be used for cost estimation of the algorithm; this
          requires latency estimation of inter-node connections as well.
    """

    def __init__(self, clocks: int = None, time: int = None, variance=0):
        # either a fixed number of clocks or a time, +- variance
        if clocks and time:
            raise ValueError(
                "cannot specify both a clock and time based latency"
            )
        self._clocks = clocks
        self._time = time  # as ns
        self.variance = variance

    def __str__(self) -> str:
        if self._clocks is not None:
            return f"c{self._clocks}"

        if self._time is not None:
            return f"t{self._time}"

        return "t0"



import unittest

import deltalanguage as dl


class MigenNodeTemplateTest(unittest.TestCase):
    """"Tests only relevant for MigenNodeTemplate."""

    def test_optional_input(self):
        """Only optional inputs are supported."""

        class AMigenNodeA(dl.MigenNodeTemplate):
            def migen_body(self, template):
                template.add_pa_in_port('in1', int)

        class AMigenNodeB(dl.MigenNodeTemplate):
            def migen_body(self, template):
                template.add_pa_in_port('in1', dl.DOptional(int))

        with self.assertRaises(dl.data_types.DeltaTypeError):
            AMigenNodeA()

        AMigenNodeB()

if __name__ == "__main__":
    unittest.main()



from __future__ import annotations
import glob
import os
import tempfile
from typing import TYPE_CHECKING, List, Tuple, Union
import zipfile

import capnp

# use the import magic from capnp to get the schema
import deltalanguage.data_types.dotdf_capnp \
    as dotdf_capnp  # pylint: disable=E0401, disable=E0611

if TYPE_CHECKING:
    from deltalanguage.wiring import DeltaGraph


def serialize_graph(
    graph: DeltaGraph,
    name: str = None,
    files: List[str] = [],
    requirements: List[str] = []
) -> Tuple[bytes, capnp.lib.capnp._DynamicStructBuilder]:
    """Converts a complete representation of the Deltaflow program stored as
    a :py:class:`DeltaGraph<deltalanguage.wiring.DeltaGraph>` to bytecode.

    Parameters
    ----------
    graph : DeltaGraph
        Deltaflow graph.
    name : str
        Name of the program. If no name is provided then this is the same
        as the name of the graph.
    files : List[str]
        Additional files required for the program to run, such as user-defined
        packages or data. Each list item can be a fixed string or a pattern.
        Duplicate filenames will only be serialised once.
    requirements : List[str]
        Additional Python packages required to execute the program. These
        packages should be freely available on PyPI and provided using their
        pip installation identifiers.

    Returns
    -------
    bytes
        Bytestring formatted according to the Deltaflow schema.
    capnp.lib.capnp._DynamicStructBuilder
        Data schema.
    """
    schema = dotdf_capnp.Program.new_message()
    if name:
        schema.name = name
    elif graph.name:
        schema.name = graph.name

    # Gather filenames, removing duplicates
    file_list = set([file for pattern in files for file in glob.glob(pattern)])
    if file_list:
        # Create zip file in temporary directory and add it to the .df file
        with tempfile.TemporaryDirectory() as zipdir:
            filename = os.path.join(zipdir, "df_zip.zip")
            with zipfile.ZipFile(filename, "w") as df_zip:
                for file in file_list:
                    df_zip.write(file)
            with open(filename, "rb") as df_zip:
                schema.files = df_zip.read()

    requirements = set(requirements)
    req_list = schema.init("requirements", len(requirements))
    for i, req in enumerate(requirements):
        req_list[i] = req

    graph.do_automatic_splitting()

    bodies = schema.init_resizable_list("bodies")
    nodes = schema.init("nodes", len(graph.nodes))
    wiring = schema.init_resizable_list("graph")

    for graph_node, capnp_node in zip(graph.nodes, nodes):
        graph_node.capnp(capnp_node, bodies)

    for graph_node in graph.nodes:
        graph_node.capnp_wiring(nodes, wiring)

    # resizable lists initialize only here
    bodies.finish()
    wiring.finish()

    return schema.to_bytes(), schema


def deserialize_graph(
    data: bytes
) -> Union[capnp._DynamicStructReader, capnp._DynamicStructBuilder]:
    """Converts bytecode to a complete graph representation of the Deltaflow
    program.

    Parameters
    ----------
    data : bytes
        Bytestring formatted according to the Deltaflow schema.

    Returns
    -------
    Union[capnp._DynamicStructReader, capnp._DynamicStructBuilder]
        Deltaflow graph.
    """
    return dotdf_capnp.Program.from_bytes(data)



"""Various utilities for internal usage."""

from typing import NamedTuple, Optional


class NamespacedName(NamedTuple):
    """A name that uses namespace, so is therefore split into
    domain and index in that domain.

    For example, it's used in :py:class:`OutPort<deltalanguage.wiring.OutPort>`
    a node has forked outputs sent to multiple destinations.
    `domain` specifies the name of the node and `n_index` specifies which
    result :py:class:`OutPort<deltalanguage.wiring.OutPort>` is meant to send
    on.
    In case if the node has a single output `n_index` is `None`.
    """

    domain: str  #: The domain name.
    n_index: Optional[str]  #: The domain index.

    def __str__(self):
        if self.n_index is None:
            return str(self.domain)
        else:
            return f"{self.domain}.{self.n_index}"


class QueueMessage():
    """Wrapper that carries messages between nodes in
    :py:class:`DeltaPySimulator<deltalanguage.runtime.DeltaPySimulator>` and
    for logging with :py:class:`MessageLog<deltalanguage.logging.MessageLog>`.

    Parameters
    ----------
    msg : object
        The message.
    clk : int
        The logical clock value when this message was created. See
        :py:class:`MessageLog<deltalanguage.logging.MessageLog>` for detail.


    .. todo:: This class belongs to the domain of `DeltaPySimulator`.
    """

    def __init__(self, msg: object, clk: int = 0):
        self.msg = msg
        self.clk = clk

    def __eq__(self, other):
        """Only compare on value, not age for normal equality."""
        if type(other) is QueueMessage:
            return self.msg == other.msg
        else:
            return False

    def __hash__(self):
        return hash(self.msg)

    def __str__(self):
        return f"{self.msg} at logical time {self.clk}"



import unittest

from deltalanguage._utils import NamespacedName, QueueMessage


class NamespacedNameTest(unittest.TestCase):

    def test_str(self):
        self.assertEqual(str(NamespacedName("a", "b")), "a.b")
        self.assertEqual(str(NamespacedName("a", None)), "a")

        with self.assertRaises(TypeError):
            NamespacedName("a")

        with self.assertRaises(TypeError):
            NamespacedName()

    def test_eq(self):
        self.assertEqual(NamespacedName("a", "b"),
                         NamespacedName("a", "b"))

        self.assertEqual(NamespacedName("a", None),
                         NamespacedName("a", None))

        self.assertNotEqual(NamespacedName("a", "b"),
                            NamespacedName("a", None))


class QueueMessageTest(unittest.TestCase):

    def test_eq(self):
        self.assertEqual(QueueMessage("test", 5),
                         QueueMessage("test", 5))

        self.assertEqual(QueueMessage("test", 5),
                         QueueMessage("test", 1))

        self.assertNotEqual(QueueMessage("test1", 6),
                            QueueMessage("test2", 6))

        self.assertEqual(hash(QueueMessage("blah-blah", 6)),
                         hash("blah-blah"))


if __name__ == "__main__":
    unittest.main()



"""Testing serialization of DeltaGraph and its componenents."""

import glob
import json
import os
import tempfile
import unittest
from unittest.mock import Mock
import zipfile
from collections import OrderedDict

import capnp
import dill

import deltalanguage.data_types.dotdf_capnp \
    as dotdf_capnp  # pylint: disable=E0401, disable=E0611
from deltalanguage._utils import NamespacedName
from deltalanguage.data_types import (BaseDeltaType,
                                      DOptional,
                                      DInt,
                                      Void,
                                      as_delta_type)
from deltalanguage.lib import StateSaver
from deltalanguage.runtime import (DeltaRuntimeExit,
                                   deserialize_graph,
                                   serialize_graph)
from deltalanguage.wiring import (DeltaBlock,
                                  DeltaMethodBlock,
                                  DeltaGraph,
                                  InPort,
                                  Interactive,
                                  MigenNodeTemplate,
                                  NodeTemplate,
                                  OutPort,
                                  PythonBody,
                                  PyConstBody,
                                  PyInteractiveBody,
                                  PyFuncBody,
                                  PyMethodBody,
                                  RealNode)
from deltalanguage.wiring._node_classes.node_bodies import PyMigenBody
from deltalanguage.wiring._node_classes.migen_node import MigenNodeTemplate

from deltalanguage.lib.hal import HardwareAbstractionLayerNode
from deltalanguage.lib.quantum_simulators import (ProjectqQuantumSimulator,
                                                  QiskitQuantumSimulator)
from test._utils import assert_capnp_content_types, return_1


class OpCacher():

    def __init__(self):
        self._add_cache = {}

    @DeltaMethodBlock()
    def cached_add(self, a: DOptional(int), b: DOptional(int)):
        if (a, b) not in self._add_cache:
            self._add_cache[(a, b)] = a+b
        raise DeltaRuntimeExit


class AMigenNode(MigenNodeTemplate):
    def migen_body(self, template):
        template.add_pa_in_port('a', DOptional(int))
        template.add_pa_in_port('b', DOptional(int))


class PortSerialisationTest(unittest.TestCase):

    def test_in_port_capnp(self):
        """Generate in port."""
        in_port = InPort(NamespacedName("node_name", "index"),
                         as_delta_type(int), None, 0)
        capnp_in_port = dotdf_capnp.InPort.new_message()
        in_port.capnp(capnp_in_port)
        self.assertEqual(capnp_in_port.name, "index")
        self.assertEqual(dill.loads(capnp_in_port.type), as_delta_type(int))
        self.assertEqual(capnp_in_port.optional, False)

    def test_in_port_capnp_optional(self):
        """Generate optional in port."""
        in_port = InPort(NamespacedName("node_name", "index"),
                         DOptional(int), None, 0)
        capnp_in_port = dotdf_capnp.InPort.new_message()
        in_port.capnp(capnp_in_port)
        self.assertEqual(capnp_in_port.name, "index")
        self.assertEqual(dill.loads(capnp_in_port.type), DInt())
        self.assertEqual(capnp_in_port.optional, True)

    def test_in_port_capnp_wiring(self):
        """Generate wiring."""
        in_port = InPort(NamespacedName("node_name", "index"),
                         as_delta_type(int), None, 0)
        wire = dotdf_capnp.Wire.new_message()
        nodes = [dotdf_capnp.Node.new_message() for _ in range(3)]
        nodes[0].name = "fake_name"
        nodes[1].name = "fake_name"
        nodes[2].name = "node_name"
        nodes[2].init("inPorts", 3)
        nodes[2].inPorts[0].name = "fake_name"
        nodes[2].inPorts[1].name = "index"
        nodes[2].inPorts[2].name = "fake_name"
        in_port.capnp_wiring(nodes, wire)
        self.assertEqual(wire.destNode, 2)
        self.assertEqual(wire.destInPort, 1)
        self.assertEqual(wire.direct, False)

    def test_in_port_capnp_wiring_direct(self):
        """In port has limit on port size."""
        in_port = InPort(NamespacedName("node_name", "index"),
                         as_delta_type(int), None, 1)
        wire = dotdf_capnp.Wire.new_message()
        nodes = [dotdf_capnp.Node.new_message()]
        nodes[0].name = "node_name"
        nodes[0].init("inPorts", 1)
        nodes[0].inPorts[0].name = "index"
        in_port.capnp_wiring(nodes, wire)
        self.assertEqual(wire.destNode, 0)
        self.assertEqual(wire.destInPort, 0)
        self.assertEqual(wire.direct, True)

    def test_out_port_capnp(self):
        """Generate out port."""
        out_port = OutPort(NamespacedName("node_name", "index"),
                           as_delta_type(int), None, None)
        capnp_out_port = dotdf_capnp.OutPort.new_message()
        out_port.capnp(capnp_out_port)
        self.assertEqual(capnp_out_port.name, "index")
        self.assertEqual(dill.loads(capnp_out_port.type), as_delta_type(int))
        with self.assertRaises(AttributeError):
            dummy = capnp_out_port.optional

    def test_out_port_capnp_no_index(self):
        """Generate out port when index not provided.
        In this case, name should be empty string.
        """
        out_port = OutPort(NamespacedName("node_name", None),
                           as_delta_type(int), None, None)
        capnp_out_port = dotdf_capnp.OutPort.new_message()
        out_port.capnp(capnp_out_port)
        self.assertEqual(capnp_out_port.name, "")
        self.assertEqual(dill.loads(capnp_out_port.type), as_delta_type(int))

    def test_out_port_capnp_wiring(self):
        """Generate wiring. Should just call same method in destination."""
        in_port = Mock()
        out_port = OutPort(NamespacedName("node_name", None),
                           as_delta_type(int), in_port, None)
        out_port.capnp_wiring([], None)
        in_port.capnp_wiring.assert_called_with([], None)


class PythonNodeSerialisationTest(unittest.TestCase):

    def setUp(self):
        DeltaGraph.clean_stack()

        @DeltaBlock(allow_const=False)
        def add_print_exit(a: DOptional(int), b: DOptional(int)) -> Void:
            print(a + b)
            raise DeltaRuntimeExit

        self.func = add_print_exit

    def test_node_serialisation(self):
        """Generate graph and check serialisation is correct."""
        with DeltaGraph() as test_graph:
            self.func(2, 3)

        _, prog = serialize_graph(test_graph)

        self.assertEqual("_".join(prog.nodes[2].name.split("_")[:-1]),
                         "add_print_exit")

        self.assertEqual(prog.nodes[2].bodies[0], 2)

        self.assertEqual(prog.nodes[2].inPorts[0].name, "a")
        self.assertEqual(dill.loads(prog.nodes[2].inPorts[0].type),
                         as_delta_type(int))
        self.assertEqual(prog.nodes[2].inPorts[0].optional, True)

        self.assertEqual(prog.nodes[2].inPorts[1].name, "b")
        self.assertEqual(dill.loads(prog.nodes[2].inPorts[1].type),
                         as_delta_type(int))
        self.assertEqual(prog.nodes[2].inPorts[1].optional, True)

        self.assertEqual(len(prog.nodes[2].inPorts), 2)
        self.assertEqual(len(prog.nodes[2].outPorts), 0)

        self.assertEqual(prog.bodies[2].python.dillImpl,
                         test_graph.nodes[2].body.as_serialized)

        self.assertEqual(len(prog.graph), 2)

        self.assertEqual(prog.graph[0].srcNode, 0)
        self.assertEqual(prog.graph[0].srcOutPort, 0)
        self.assertEqual(prog.graph[0].destNode, 2)
        self.assertEqual(prog.graph[0].destInPort, 0)
        self.assertEqual(prog.graph[0].direct, False)

        self.assertEqual(prog.graph[1].srcNode, 1)
        self.assertEqual(prog.graph[1].srcOutPort, 0)
        self.assertEqual(prog.graph[1].destNode, 2)
        self.assertEqual(prog.graph[1].destInPort, 1)
        self.assertEqual(prog.graph[1].direct, False)

    def test_node_serialisation_duplicate_bodies(self):
        """If two blocks share the same body only keep one copy."""
        with DeltaGraph() as test_graph:
            self.func(2, 3)
            self.func(4, 5)

        _, prog = serialize_graph(test_graph)

        self.assertEqual(len(prog.bodies), 5)
        self.assertEqual(prog.nodes[2].bodies[0], prog.nodes[5].bodies[0])

    def test_node_serialisation_multi_body_node(self):
        """If two blocks share the same body only keep one copy."""
        with DeltaGraph() as test_graph:
            n1 = self.func(2, 3)

        @DeltaBlock(allow_const=False)
        def over_complex_add(a: DOptional(int), b: DOptional(int)):
            raise DeltaRuntimeExit

        @Interactive(in_params=OrderedDict([('a', DOptional(int)),
                                            ('b', DOptional(int))]))
        def broken_adder(node: RealNode):
            node.receive('a')
            node.receive('b')
            raise DeltaRuntimeExit

        n1.add_body(AMigenNode())
        n1.add_body(over_complex_add)
        n1.add_body(OpCacher().cached_add)
        n1.add_body(broken_adder)

        _, prog = serialize_graph(test_graph)

        self.assertEqual(len(prog.nodes[2].bodies), 5)

    def test_splitter_serialisation(self):
        """Splitter nodes should be added when serialising."""
        @DeltaBlock(allow_const=False)
        def add(a: int, b: int) -> int:
            return a+b
        with DeltaGraph() as test_graph:
            a = add(2, 3)
            self.func(a, 4)
            self.func(a, 5)

        _, prog = serialize_graph(test_graph)

        splitter_body = dill.loads(prog.bodies[-1].python.dillImpl)
        self.assertEqual(type(splitter_body), PyFuncBody)
        self.assertEqual(len(prog.nodes), 8)

    def test_interactive_serialisation(self):
        """Interactive nodes have their own serialised body type.

        This is to distinguish their bodies from Python bodies, as they
        take different inputs.
        """
        @Interactive(in_params={'a': int, 'b': int}, out_type=Void)
        def add(node) -> int:
            a = node.receive('a')
            b = node.receive('b')
            print(a+b)
            raise DeltaRuntimeExit

        with DeltaGraph() as test_graph:
            a = add.call(a=2, b=3)

        _, prog = serialize_graph(test_graph)

        self.assertEqual(prog.bodies[2].which(), 'interactive')
        interactive_body = prog.bodies[2].interactive.dillImpl
        self.assertEqual(interactive_body, a.body.as_serialized)

    def test_projectQ_serialisation(self):
        """Test ProjectQ nodes serialization/deserialization.

        ProjectQ can't be fully serialized, we need to exclude from dill the
        engine (C++ libraries). This test is to guarantee that when we
        deserialize everything works as expected.
        """

        with DeltaGraph() as test_graph:
            HardwareAbstractionLayerNode(
                ProjectqQuantumSimulator(register_size=2)).accept_command(
                    hal_command=0x4000000)

        data, _ = serialize_graph(test_graph)
        g_capnp = deserialize_graph(data)

        # Checking that we are investigating the right node.
        self.assertEqual(g_capnp.nodes[1].name.split("_")[0], "accept")
        body = g_capnp.bodies[1].python.dillImpl

        node = dill.loads(body)
        node.eval(hal_command=0x4000000)

    def test_qiskit_serialisation(self):
        """Test Qiskit nodes serialization/deserialization.
        """

        with DeltaGraph() as test_graph:
            HardwareAbstractionLayerNode(
                QiskitQuantumSimulator(register_size=2)
            ).accept_command(hal_command=0x4000000)

        data, _ = serialize_graph(test_graph)
        g_capnp = deserialize_graph(data)

        # Checking that we are investigating the right node.
        self.assertEqual(g_capnp.nodes[1].name.split("_")[0], "accept")
        body = g_capnp.bodies[1].python.dillImpl
        node = dill.loads(body)
        node.eval(hal_command=0x4000000)

    def test_template_node_capnp(self):
        """Test serialisation of nodes with no body.
        """
        template = NodeTemplate(in_params={'a': int, 'b': int},
                                out_type=int, name="temp-test")

        with DeltaGraph() as test_graph:
            template.call(a=1, b=2)

        data, _ = serialize_graph(test_graph)
        g_capnp = deserialize_graph(data)

        for n in g_capnp.nodes:
            if n.name.split("_")[1] == 'temp-test':
                node = n
                break
        self.assertEqual(node.name.split("_")[0], 'template')
        self.assertEqual(len(node.bodies), 0)

    def test_serialisation(self):
        """Serialize/deserialize a graph.

        .. note::
            The content of the bodies depends on the environment, i.e. how
            the test is executed.
            For this reason we just compare the structure of the graph here.
        """
        with DeltaGraph() as graph:
            self.func(40, 2)

        data, _ = serialize_graph(graph)
        self.assertEqual(type(data), bytes)
        g_capnp = deserialize_graph(data).to_dict()
        assert_capnp_content_types(self, g_capnp)
        with open('test/data/graph_capnp.json', 'r') as file:
            self.assertEqual(g_capnp, json.load(file))

    def test_multi_body_serialisation(self):
        """Tests a graph with a multi-body node is serialised and matches
        a target capnp file.
        """
        with DeltaGraph() as graph:
            n1 = self.func(40, 2)

        @DeltaBlock(allow_const=False)
        def simple_add_2(a: DOptional(int), b: DOptional(int)):
            raise DeltaRuntimeExit

        n1.add_body(simple_add_2)

        data, _ = serialize_graph(graph)
        self.assertEqual(type(data), bytes)
        g_capnp = deserialize_graph(data).to_dict()
        assert_capnp_content_types(self, g_capnp)
        with open('test/data/graph_multibody_capnp.json', 'r') as file:
            self.assertEqual(g_capnp, json.load(file))


class FileSerialisationTest(unittest.TestCase):
    """Tests of serialised files."""

    def setUp(self):
        """Set up a simple graph"""
        saver = StateSaver()
        with DeltaGraph() as test_graph:
            saver.save_and_exit(return_1())
        self.graph = test_graph

    def assert_correct_file_serialisation(self, files):
        _, prog = serialize_graph(self.graph, files=files)
        if len(files) == 0:
            self.assertEqual(prog.files, b'')
        else:
            with tempfile.TemporaryDirectory() as zip_dir:
                filename = os.path.join(zip_dir, "df_zip.zip")
                with open(filename, "wb") as df_zip:
                    df_zip.write(prog.files)
                with zipfile.ZipFile(filename, "r") as df_zip:
                    file_list = set([file
                                     for pattern in files
                                     for file in glob.glob(pattern)])
                    self.assertEqual(len(file_list), len(df_zip.namelist()))
                    for file in file_list:
                        self.assertIn(file, df_zip.namelist())
                        with open(file, "rb") as content:
                            self.assertEqual(content.read(), df_zip.read(file))

    def assert_correct_reqs_serialisation(self, reqs):
        _, prog = serialize_graph(self.graph, requirements=reqs)
        reqs = set(reqs)
        self.assertEqual(len(prog.requirements), len(reqs))
        for requirement in reqs:
            self.assertIn(requirement, prog.requirements)

    def test_serialisation_no_file(self):
        """Generate serialisation without file attached."""
        self.assert_correct_file_serialisation([])

    def test_serialisation_with_file(self):
        """Generate serialisation with file attached."""
        files = [os.path.join("deltalanguage", "lib", "primitives.py")]
        self.assert_correct_file_serialisation(files)

    def test_serialisation_multiple_files(self):
        """Generate serialisation with several files attached."""
        files = [os.path.join("deltalanguage", "lib", "primitives.py"),
                 os.path.join("deltalanguage", "runtime", "_output.py"),
                 os.path.join("deltalanguage", "__about__.py")]
        self.assert_correct_file_serialisation(files)

    def test_serialisation_nonpy_files(self):
        """Generate serialisation with non-Python files attached."""
        files = [os.path.join("deltalanguage", "DeltaStyle.xml"),
                 os.path.join("test", "data", "graph_capnp.json"),
                 os.path.join("examples", "tutorials", "const_nodes.ipynb"),
                 os.path.join("docs", "figs", "blocks.png")]
        self.assert_correct_file_serialisation(files)

    def test_serialisation_duplicate_files(self):
        """Remove duplicate files from serialisation."""
        files = [os.path.join("deltalanguage", "DeltaStyle.xml"),
                 os.path.join("deltalanguage", "DeltaStyle.xml")]
        self.assert_correct_file_serialisation(files)

    def test_serialisation_pattern(self):
        """Generate serialisation with file patterns."""
        files = [os.path.join("deltalanguage", "*", "__init__.py")]
        self.assert_correct_file_serialisation(files)

    def test_serialisation_no_reqs(self):
        """Generate serialisation without requirements."""
        self.assert_correct_reqs_serialisation([])

    def test_serialisation_one_reqs(self):
        """Generate serialisation with one requirement."""
        self.assert_correct_reqs_serialisation(["numpy"])

    def test_serialisation_several_reqs(self):
        """Generate serialisation with multiple requirements."""
        self.assert_correct_reqs_serialisation(
            ["numpy", "matplotlib", "pandas"])

    def test_serialisation_duplicate_reqs(self):
        """Each requirement only included once."""
        self.assert_correct_reqs_serialisation(["numpy", "numpy", "pandas"])

    def test_serialisation_version_reqs(self):
        """Version numbers can be included in requirements."""
        self.assert_correct_reqs_serialisation(
            ["numpy==1.20.0", "matplotlib>=3.3"])


if __name__ == "__main__":
    unittest.main()



from enum import Enum
from typing import Iterable, List

from deltalanguage.data_types import DUInt, DSize, DArray

# Command fields
# 31-26       25-16        15-0
# [OPCODE]    [ARGUMENT]   [QUBIT INDEX]


class Opcode(Enum):
    """Operational codes for HAL."""

    NOP = 0
    STATE_PREPARATION = 1
    STATE_MEASURE = 2
    SEND_TO_HOST = 3
    STOP = 4

    H = 5
    R = 6
    RX = 7
    RY = 8
    RZ = 9
    S = 10
    SQRT_X = 11
    T = 12
    X = 13
    Y = 14
    Z = 15

    CONTROL = 16

    INVS = 17

    ID = 18  # identity gate
    SX = 19  # pi-rotation around x+y (required for randomized compiling)
    SY = 20  # pi-rotation around x-y (required for randomized compiling)

    # pi-rotations with axes in the xy, yz, and zx-planes, respectively
    PIXY = 21
    PIYZ = 22
    PIZX = 23


class Shifts(Enum):
    """
    .. TODO:: Missing description.
    """

    ARG = 16
    OPCODE = 26
    VALIDS = 16


class Masks(Enum):
    """
    .. TODO:: Missing description.
    """

    QUBIT_INDEX = 65535
    ARG = 67043328
    VALIDS = 4294901760
    MEASUREMENTS = 65535


def command_creator(op: str, argument=0, qubit=0) -> DUInt(DSize(32)):
    """Helper function to create HAL commands.

    Parameters
    ----------
    op : str
        Name of opcode.
    argument : int
        Integer representation of argument value
    qubit : int
        Integer representation of qubit address

    Returns
    -------
    DUInt(DSize(32))
        32-bit HAL command
    """
    return (Opcode[op].value << Shifts.OPCODE.value) \
        | (argument << Shifts.ARG.value) | qubit


def command_unpacker(cmd: DUInt(DSize(32))):
    """Helper function to unpack HAL commands.

    Parameters
    ----------
    cmd : DUInt(DSize(32))
        32-bit HAL command.

    Returns
    -------
    op : str
        Name of opcode.
    argument : int
        Integer representation of argument value.
    qubit : int
        Integer representation of qubit address.
    """
    op_mask = sum(map(lambda n: 2**n, range(Shifts.OPCODE.value, 32)))
    op = (cmd & op_mask) >> Shifts.OPCODE.value
    # We pass strings around rather than Opcode elements, so we use
    # this to get the reverse mapping from the Opcode class
    # This is semi-safe, as we don't have repeated names for operations
    op = Opcode._value2member_map_[op]._name_

    arg_mask = sum(map(lambda n: 2**n,
                       range(Shifts.ARG.value, Shifts.OPCODE.value)))
    arg = (cmd & arg_mask) >> Shifts.ARG.value

    qubit_mask = sum(map(lambda n: 2**n, range(0, Shifts.VALIDS.value)))
    qubit = (cmd & qubit_mask)
    return (op, arg, qubit)


def measurement_unpacker(bitcode: DUInt(DSize(32)), qubits: Iterable) -> List:
    """Helper function to convert 32-bit status result from HAL into an array
    of measurements for given qubit indices.

    Parameters
    ----------
    bitcode : DUInt(DSize(32))
        32-bit measurement status from HAL.
    qubits : Iterable
        List of qubits for which the measurement result will be returned.

    Returns
    -------
    List
        List of measurement results for the specified qubits.

    Raises
    ------
    ValueError
        If not all valid flags are 1.
    """

    # split status bitcode into measurements (first 16 bits) and
    # valid flags (last 16 bits)
    measurements = bitcode & Masks.MEASUREMENTS.value
    valids = (bitcode & Masks.VALIDS.value) << Shifts.VALIDS.value

    # print('1', measurements)

    if valids != Masks.VALIDS.value << Shifts.VALIDS.value:
        raise ValueError(f"Invalid measurement!, {hex(valids)}, "
                         f"{hex(Masks.VALIDS.value << Shifts.VALIDS.value)}")

    measurements_list = []

    for i in qubits:
        measurements_list.append((measurements >> i) & 1)

    # print('2', measurements_list)
    return measurements_list


def measurement_creator(measurement_array, qubits):
    """Helper function to convert 32-bit status result from HAL into an array
    of measurements for given qubit indices.

    Parameters
    ----------
    measurement_array
        List of measurement results for the specified qubits.
    qubits : Iterable
        List of qubits for which the measurement result will be returned.

    Returns
    -------
    bitcode : DUInt(DSize(32))
        32-bit measurement status from HAL.
    """
    packed_measurements = 0
    for i in qubits:
        measurement = measurement_array.pop()
        packed_measurements |= 2**i if measurement else 0

    packed_measurements |= Masks.VALIDS.value

    return packed_measurements



"""
The standard library for the Deltaflow language contains:

- Primitives for generation, re-routing, and collection of data:
  :py:func:`make_generator`
  :py:func:`make_splitter`
  :py:class:`StateSaver`

- Routines for the Hardware Abstraction Layer (HAL) and Quantum Simulators,
  which are best described in :doc:`examples/examples`.
"""

from deltalanguage.lib.hal import (HardwareAbstractionLayerNode,
                                   Masks,
                                   Opcode,
                                   Shifts,
                                   command_creator,
                                   measurement_unpacker,
                                   hal_template)
from deltalanguage.lib.primitives import (make_generator,
                                          make_splitter,
                                          StateSaver)
from deltalanguage.lib.quantum_simulators import (IQuantumSimulator,
                                                  ProjectqQuantumSimulator,
                                                  QiskitQuantumSimulator)



from ._interface_quantum_simulator import IQuantumSimulator
from ._projectq_quantum_simulator import ProjectqQuantumSimulator
from ._qiskit_quantum_simulator import QiskitQuantumSimulator
from ._qiskit_noise_model import CustomNoiseModel



import logging
import threading
import time
import unittest

from test._utils import (TwoInts,
                         TwoIntsT,
                         add_non_const,
                         const_consumer,
                         return_1,
                         return_2)

from deltalanguage.data_types import as_delta_type, Void, DeltaIOError
from deltalanguage.lib import StateSaver
from deltalanguage.runtime import DeltaPySimulator, DeltaRuntimeExit
from deltalanguage._utils import QueueMessage
from deltalanguage.wiring import (DeltaBlock,
                                  DeltaGraph,
                                  Interactive,
                                  PythonNode,
                                  placeholder_node_factory)


@DeltaBlock(name="node")
def int_to_str(n: int) -> str:
    return str(n)


@Interactive(in_params={"num": str}, out_type=int, name="blah")
def interactive_func(node: PythonNode):
    node.send(3)
    num = node.receive()["num"]
    num = int(num, base=10)
    node.send(num + 1)


@Interactive(in_params={"num": int}, out_type=int)
def forward(node: PythonNode):
    for _ in range(10):
        num = node.receive("num")
        node.send(num)
    raise DeltaRuntimeExit


@Interactive(in_params={'num': int}, out_type=TwoIntsT)
def add_until_10(node: PythonNode):
    """Sends num on the left port until it is greater than 10, then sends it
    on the right port.
    """
    num = 1
    while num < 10:
        node.send(TwoInts(num, None))
        num = node.receive()["num"]
    node.send(TwoInts(None, num))


class InteractiveNodeGeneralTest(unittest.TestCase):
    """Unsorted testes of interactive nodes."""

    def test_interactive_node_without_io(self):
        """Interactive node must have I/O. 

        We chose this rule because runtime simulators,
        in particular the Python GIL, wouldn't know when to
        interrupt this node simulation.

        In order to allow such nodes we need to change the simulation strategy.
        """

        with self.assertRaises(DeltaIOError):
            @Interactive({}, Void)
            def bar(node):
                while True:
                    pass

    def test_interactive_node_without_ports(self):
        """An interactive node with I/O can be just forgotten during graph
        construction. We check that it is connected.
        """
        @Interactive({}, int)
        def bar(node):
            node.send(1)

        with DeltaGraph() as graph:
            bar.call()

        with self.assertRaises(DeltaIOError):
            graph.check()


class SimpleGraph(unittest.TestCase):

    def setUp(self):
        with DeltaGraph() as graph:
            ident_ph = placeholder_node_factory()
            inter = interactive_func.call(num=ident_ph)
            ident = int_to_str(inter)
            ident_ph.specify_by_node(ident)
        self.graph = graph
        self.runtime = DeltaPySimulator(self.graph)
        self.dummy_node = ident
        self.inter_node = inter

    def test_ports(self):
        """Test if the node wires are connected correctly and have the right
        type.
        """
        interactive_node = self.graph.find_node_by_name("blah")
        dummy_node = self.graph.find_node_by_name("node")

        self.assertEqual(len(interactive_node.in_ports), 1)
        self.assertEqual(len(interactive_node.out_ports), 1)
        self.assertTupleEqual(interactive_node.out_ports[0].port_name,
                              (interactive_node.name, None))
        self.assertEqual(str(interactive_node.out_ports[0].port_type),
                         str(as_delta_type(int)))

        self.assertEqual(len(dummy_node.in_ports), 1)
        self.assertEqual(len(dummy_node.out_ports), 1)
        self.assertTupleEqual(interactive_node.out_ports[0].port_name,
                              (interactive_node.name, None))
        self.assertEqual(str(dummy_node.out_ports[0].port_type),
                         str(as_delta_type(str)))

        self.assertEqual(interactive_node.out_ports[0].destination.node,
                         dummy_node)

    def test_simulated_run(self):
        """Test if the Interactive Node performs correctly by simulating
        the runtime and the other node.
        """
        inter_node: PythonNode = self.graph.find_node_by_name("blah")
        in_q = self.runtime.in_queues[inter_node.name]["num"]
        out_q = self.runtime.out_queues[inter_node.name][None]

        thread = threading.Thread(target=inter_node.thread_worker,
                                  args=(self.runtime,))
        thread.start()

        self.assertEqual(out_q.get().msg, 3)
        in_q.put(QueueMessage("3"))
        self.assertEqual(out_q.get().msg, 4)

        thread.join()
        self.assertFalse(thread.is_alive())

    def test_real_run(self):
        """Test that the graph runs and terminates."""
        interactive_node = self.graph.find_node_by_name("blah")

        runtime = DeltaPySimulator(self.graph)
        # run for a short time - just enough for the processing to end
        runtime.run(0.1)
        # check that the interactive node has terminated
        self.assertFalse(runtime.threads[interactive_node.name].is_alive())
        runtime.stop()

    def test_graph_properties(self):
        self.graph.check()


class UpdateClockTest(unittest.TestCase):
    """Test logical clock in MessageLog.

    TODO this test case should be moved from here.
    """

    def setUp(self):
        with DeltaGraph() as self.test_graph:
            const_consumer(obj=forward.call(num=return_1()))

    def test_clock_updates(self):
        """Messages should be displayed at end of run in logical clock order.
        """
        self.rt = DeltaPySimulator(self.test_graph, msg_lvl=logging.DEBUG)
        self.rt.run()
        message_times = [msg.clk for node, _, msg in self.rt.msg_log.messages
                         if node.startswith("const_consumer")]
        self.assertGreater(len(message_times), 0)
        for time_1, time_2 in zip(message_times, message_times[1:]):
            self.assertLess(time_1, time_2)

    def test_no_message_logging(self):
        """Messages should not be logged at all.
        """
        self.rt = DeltaPySimulator(self.test_graph)
        self.rt.run()
        message_times = [msg.clk for node, _, msg in self.rt.msg_log.messages
                         if node.startswith("const_consumer")]
        self.assertEqual(message_times, [])
        for time_1, time_2 in zip(message_times, message_times[1:]):
            self.assertLess(time_1, time_2)


class ComplexGraph(unittest.TestCase):

    def setUp(self):
        """Constructs the following graph:
        +---------------------------+
        |             SAVE          |
        |   +-----+    ^            |
        +--->Add  +----+            |
            |to 10|        +----+   |
            |     +-------->    |   |
            +-----+        |ADD +---+
                     2 --->|    |
                           +----+
        """
        self.saver = StateSaver(int)

        with DeltaGraph() as graph:
            add_ph = placeholder_node_factory()
            b = return_2()
            self.int_node = add_until_10.call(num=add_ph)
            add_node = add_non_const(b, self.int_node.x)
            add_ph.specify_by_node(add_node)
            self.saver.save_and_exit(self.int_node.y)
        self.graph = graph
        self.runtime = DeltaPySimulator(self.graph)

    def test_run(self):
        """This graph should run, adding 2 to the number on every cycle."""
        self.runtime.run()
        self.assertEqual(self.saver.saved, [11])

    def test_graph_properties(self):
        self.graph.check()


if __name__ == "__main__":
    unittest.main()



"""Testing InPort and OutPort classes."""

import unittest

from deltalanguage.data_types import DFloat, DInt, DOptional, DUnion
from deltalanguage._utils import NamespacedName
from deltalanguage.wiring import InPort, OutPort


class TestInPortDOptiona(unittest.TestCase):
    """This list of test shows that InPort accepts the expected data type
    when created and does the following simplification:
    if it's `DOptional(t)`, the port type becomes just `t`.
    """

    def test_non_DOptional(self):
        port = InPort(NamespacedName("test_name", None),
                      DInt(),
                      None,
                      0)
        self.assertEqual(port.port_type, DInt())
        self.assertEqual(port.is_optional, False)

    def test_DOptional(self):
        port = InPort(NamespacedName("test_name", None),
                      DOptional(DInt()),
                      None,
                      0)
        self.assertEqual(port.port_type, DInt())
        self.assertEqual(port.is_optional, True)

    def test_DUnion_of_single(self):
        """DUnion of a single type is not converted to a single type."""
        port = InPort(NamespacedName("test_name", None),
                      DUnion([DInt()]),
                      None,
                      0)
        self.assertEqual(port.port_type, DUnion([DInt()]))
        self.assertEqual(port.is_optional, False)

    def test_DUnion(self):
        port = InPort(NamespacedName("test_name", None),
                      DUnion([DInt(), DFloat()]),
                      None,
                      0)
        self.assertEqual(port.port_type, DUnion([DInt(), DFloat()]))
        self.assertEqual(port.is_optional, False)

    def test_DOptiona_of_DUnion(self):
        port = InPort(NamespacedName("test_name", None),
                      DOptional(DUnion([DInt(), DFloat()])),
                      None,
                      0)
        self.assertEqual(port.port_type, DUnion([DInt(), DFloat()]))
        self.assertEqual(port.is_optional, True)


class TestOutPortDOptiona(unittest.TestCase):
    """This list of test shows that OutPort data type cannot be `DOptional`."""

    def test_non_DOptional(self):
        port = OutPort(NamespacedName("test_name", None),
                       DInt(),
                       None,
                       None)
        self.assertEqual(port.port_type, DInt())

    def test_DOptional(self):
        port = OutPort(NamespacedName("test_name", None),
                       DOptional(DInt()),
                       None,
                       None)
        with self.assertRaises(TypeError):
            dummy = port.port_type


if __name__ == "__main__":
    unittest.main()



import logging
from typing import TYPE_CHECKING, Callable, Dict, Union, Optional, Type

from deltalanguage.data_types import (as_delta_type,
                                      BaseDeltaType,
                                      DeltaIOError,
                                      DOptional,
                                      Void)
from deltalanguage.logging import make_logger
from ._node_classes.latency import Latency
from ._node_classes.node_bodies import Body
from ._node_classes.real_nodes import (PythonNode,
                                       as_node,
                                       get_func_in_params_out_type,
                                       in_params_as_delta_types)
from ._body_templates import (BodyTemplate,
                              FuncBodyTemplate,
                              InteractiveBodyTemplate,
                              MethodBodyTemplate)

if TYPE_CHECKING:
    from ._delta_graph import DeltaGraph

log = make_logger(logging.WARNING, "Node Templates")


class NodeTemplate():
    """NodeTemplates are recipes for how to create a specific node.
    They have a list of :py:class:`BodyTemplate`s that represent how to
    construct a number of different bodies for the node. Each
    ``BodyTemplate`` has a call method to allow creation of the
    node. This node will be given bodies from all valid ``BodyTemplates``.

    Parameters
    ----------
    name : str
        Default name for the nodes created using this node template
    in_params : Dict[str, Union[BaseDeltaType, DOptional]]
        List of input parameters for nodes created using this template
    out_type : Type
        Out type for nodes created using this template
    node_key : Optional[str]
        Node key to allow node object to be accessed inside bodies
        created using this template
    in_port_size : int
        Bounded size of port queues into this node
    """

    def __init__(self,
                 in_params: Dict[str, Type],
                 out_type: Type = Void,
                 name: str = None,
                 lvl: int = logging.ERROR,
                 node_key: Optional[str] = None,
                 in_port_size: int = 0):
        self.name = "template_" + name if name else "template"
        self.lvl = lvl
        self.node_key = node_key
        self.in_port_size = in_port_size

        self._body_templates = []

        self.in_params = in_params_as_delta_types(in_params)
        self.out_type = as_delta_type(out_type)

        self._has_optional_inputs = False
        for in_type in in_params.values():
            if isinstance(in_type, DOptional):
                self._has_optional_inputs = True

    def _standardised_call(self, graph, name: str, lvl: int,
                           caller: Optional[BodyTemplate] = None,
                           body: Optional[Body] = None,
                           *nodes, **kwnodes):
        """Internal method used by different node creation methods
        Contains several common behaviors for node creation

        Parameters
        ----------
        caller : BodyTemplate
            The BodyTemplate that initiated this node creation
        graph : DeltaGraph
            The graph this node will be attached to
        body : Body
            The body the caller BodyTemplate has provided for node creation
            This body is special as it always goes first in the body list.
        name : str
            Name of the node to create
        """
        # Input node sanitisation
        pos_in_nodes = [as_node(arg, graph) for arg in nodes]
        kw_in_nodes = {param_name: as_node(arg, graph)
                       for (param_name, arg) in kwnodes.items()}

        # Const folding logic
        all_args = pos_in_nodes + list(kw_in_nodes.values())
        num_const = sum([node.is_const() for node in all_args])
        all_allow_const = all(st.allow_const for st in self._body_templates)

        # Condition for node to be created as constant (all const bodies)
        const_condition_met = False
        if (all_allow_const
            and num_const == len(pos_in_nodes) + len(kw_in_nodes)
            and self.node_key is None
                and not self._has_optional_inputs):
            const_condition_met = True

        # Creation of bodies for all BodyTemplates
        if caller and const_condition_met:
            body = caller.construct_const_body(*pos_in_nodes, **kw_in_nodes)

        bodies = []
        if body is not None:
            bodies.append(body)

        for body_template in self._body_templates:
            if (body_template.construction_ready
                    and body_template is not caller):
                if const_condition_met:
                    bodies.append(body_template.construct_const_body(
                        *pos_in_nodes,
                        **kw_in_nodes))
                else:
                    bodies.append(body_template.construct_body())

        # Node creation and return
        ret_node = PythonNode(graph,
                              bodies,
                              self.in_params,
                              pos_in_nodes,
                              kw_in_nodes,
                              out_type=self.out_type,
                              name=name,
                              in_port_size=self.in_port_size,
                              node_key=self.node_key,
                              lvl=lvl)
        log.debug(f"making {type(ret_node)}: {ret_node.name}")
        return ret_node

    def merge(self, other: 'NodeTemplate'):
        """Check compatibility of two ``NodeTemplates`` and then merge the 
        one given as a parameter into this one.

        Parameters
        ----------
        other : NodeTemplate
            Other ``NodeTemplate`` to try to merge into this one.
        """
        if other is not self:
            if other.compatible(self.node_key, self.in_port_size,
                                self.in_params, self.out_type):
                for other_body_t in other._body_templates:
                    other_body_t.set_node_template(self)
                    self._body_templates.append(other_body_t)

    def add_constructor(self, body_template: Union[object, BodyTemplate]):
        """Add a body constructor to this ``NodeTemplate`` via a 
        ``BodyTemplate``.

        .. note::
          Main usage of this function is for when you have some body that can
          belong on your node constructor but the definition of that body is
          in some non-accessible code.

        Parameters
        ----------
        body_template : Union[object, BodyTemplate]
            ``BodyTemplate`` to merge or some object for which
            ``template`` is defined, where ``template`` is a ``BodyTemplate``
        """
        if not isinstance(body_template, BodyTemplate):
            body_template = body_template.template

        if body_template in self._body_templates:
            log.warning(f"Body template {body_template.name} attempted to join"
                        f" node template {self.name} more than one time")
        else:
            if body_template.compatible(self.node_key, self.in_port_size,
                                        self.in_params, self.out_type):
                self._body_templates.append(body_template)
                body_template.set_node_template(self)

    @staticmethod
    def _standardised_merge(
        other: Optional['NodeTemplate'],
        body_template: BodyTemplate,
        node_key: Optional[str],
        in_port_size: int,
        in_params: Dict[str, Union[BaseDeltaType, DOptional]],
        out_type: Union[BaseDeltaType, DOptional]
    ) -> BodyTemplate:
        """Internal method used by template merging and creation methods
        Contains the common behaviour for checking other_template is compatible
        with the given information. Also the common behaviour of creating a new
        NodeTemplate when there is no other_template.

        Parameters
        ----------
        other : Optional[NodeTemplate]
            The template that this method will add the body_template to after
            checks. May be None in which case new NodeTemplate created.
        body_template : BodyTemplate
            The BodyTemplate looking to be given a NodeTemplate to merge into.

        Returns
        -------
        BodyTemplate
            Returns the given BodyTemplate with its parent NodeTemplate
            attached. This parent NodeTemplate is either ``other_template`` or
            a newly created NodeTemplate.
        """
        if other and other.compatible(node_key, in_port_size,
                                      in_params, out_type):
            main_template = other
        else:
            main_template = NodeTemplate(in_params, out_type,
                                         name=body_template.name,
                                         node_key=node_key,
                                         in_port_size=in_port_size)
        body_template.set_node_template(main_template)
        main_template._body_templates.append(body_template)
        return body_template

    def compatible(self,
                   node_key: Optional[str],
                   in_port_size: int,
                   in_params: Dict[str, Union[BaseDeltaType, DOptional]],
                   out_type: Union[BaseDeltaType, DOptional]) -> bool:
        """Checks compatibility between this ``NodeTemplate`` and some params
        that are important for node creation.

        Returns
        -------
        bool
            ``True`` if paramaters are compatible with this ``NodeTemplate``.
        """
        if self.in_params != in_params:
            raise ValueError(f"Node template {self.name} "
                             "incompatible with given input parameters.")
        if self.out_type != out_type:
            raise ValueError(f"Node template {self.name} "
                             "incompatible with given out type.")
        if self.node_key != node_key:
            raise ValueError(f"Node template {self.name} "
                             "incompatible with given node type.")
        if self.in_port_size != in_port_size:
            raise ValueError(f"Node template {self.name} "
                             "incompatible with given in port size.")
        return True

    @staticmethod
    def merge_deltablock(other: Optional['NodeTemplate'],
                         body_func: Callable,
                         allow_const: bool,
                         node_key: Optional[str] = None,
                         name: str = None,
                         in_port_size: int = 0,
                         latency: Latency = None,
                         lvl: int = logging.ERROR) -> BodyTemplate:
        """Create a :py:class:`BodyTemplate` for the bodies and constructor
        created using the ``@DeltaBlock`` decorator. Associate this with an
        existing or newly created :py:class:`NodeTemplate`.
        """
        in_params, out_type = get_func_in_params_out_type(
            body_func, False, node_key
        )

        body_template = FuncBodyTemplate(
            name, latency, lvl, body_func, allow_const)
        return NodeTemplate._standardised_merge(other,
                                                body_template,
                                                node_key,
                                                in_port_size,
                                                in_params,
                                                out_type)

    @staticmethod
    def merge_deltamethod(other: Optional['NodeTemplate'],
                          body_func: Callable,
                          node_key: Optional[str] = None,
                          name: str = None,
                          in_port_size: int = 0,
                          latency: Latency = None,
                          lvl: int = logging.ERROR) -> BodyTemplate:
        """Create a :py:class:`BodyTemplate` for the bodies and constructor
        created using the ``@DeltaMethodBlock`` decorator. Associate this with
        an existing or newly created ``NodeTemplate``.
        """
        in_params, out_type = get_func_in_params_out_type(
            body_func, True, node_key,
        )

        body_template = MethodBodyTemplate(name, latency, lvl, body_func)
        return NodeTemplate._standardised_merge(other,
                                                body_template,
                                                node_key,
                                                in_port_size,
                                                in_params,
                                                out_type)

    @staticmethod
    def merge_interactive(other: Optional['NodeTemplate'],
                          func: Callable[[PythonNode], None],
                          in_params: Dict[str, Type],
                          out_type: Type,
                          name: str = None,
                          in_port_size: int = 0,
                          latency: Latency = None,
                          lvl: int = logging.ERROR) -> BodyTemplate:
        """Create a :py:class:`BodyTemplate` for the bodies and constructor
        created using the ``@Interactive`` decorator. Associate this with an
        existing or newly created ``NodeTemplate``.
        """
        if len(in_params) == 0 and out_type == Void:
            raise DeltaIOError('Interactive node must have either an input '
                               'or an output. Otherwise it may freeze '
                               'the runtime simulator.')

        in_params = in_params_as_delta_types(in_params)
        out_type = as_delta_type(out_type)
        body_template = InteractiveBodyTemplate(name, latency, lvl, func)
        return NodeTemplate._standardised_merge(other,
                                                body_template,
                                                node_key=None,
                                                in_port_size=in_port_size,
                                                in_params=in_params,
                                                out_type=out_type)

    @staticmethod
    def merge_migenblock(other: Optional['NodeTemplate'],
                         body_template: BodyTemplate,
                         in_params: Dict[str, Type],
                         out_type: Type,) -> BodyTemplate:
        """Create a :py:class:`BodyTemplate` for the bodies and constructor
        for migen based bodies. Associate this with an existing or newly
        created :py:class:`NodeTemplate`.
        """
        in_params = in_params_as_delta_types(in_params)
        out_type = as_delta_type(out_type)

        return NodeTemplate._standardised_merge(other=other,
                                                body_template=body_template,
                                                in_params=in_params,
                                                out_type=out_type,
                                                node_key=None,
                                                in_port_size=0)

    def call(self, *args, **kwargs):
        """The default call to create a node, can be
        used when no :py:class:`BodyTemplate` exists.
        Tries to construct all bodies in order they were added to the
        template. 
        """
        from ._delta_graph import DeltaGraph
        if DeltaGraph.stack():
            graph = DeltaGraph.current_graph()
        else:
            raise ValueError("Node template called when no graph was active.")

        return self._standardised_call(graph, self.name, self.lvl, None, None,
                                       *args, **kwargs)



"""
This submodule covers the essential bits about the constraction of the
Deltaflow graph and contains the following user-facing parts:

- :py:class:`DeltaGraph` is the central part of the Deltaflow language.

- Decorators used for quick node construction:
  :py:class:`DeltaBlock`
  :py:class:`DeltaMethodBlock`
  :py:class:`Interactive`

- Wiring routines:
  :py:func:`placeholder_node_factory`

- Node classes 
  :py:class:`RealNode`
  :py:class:`PythonNode`

For convenience the classes and functions listed above are also exposed to
the top level module and can be used like this:

.. code-block:: python

    import deltalanguage as dl
    
    print(dl.DeltaGraph())
    print(dl.DeltaBlock())

The learning path for new user is best described in :doc:`tutorials/tutorials`.
"""

from ._node_classes.abstract_node import AbstractNode, ForkedNode, ProxyNode
from ._node_classes.latency import Latency
from ._node_classes.migen_node import MigenNodeTemplate
from ._node_classes.node_bodies import (PyConstBody,
                                        PyFuncBody,
                                        PyInteractiveBody,
                                        PyMethodBody,
                                        PyMigenBody,
                                        PythonBody)
from ._node_classes.placeholder_node import PlaceholderNode
from ._node_classes.port_classes import InPort, OutPort
from ._node_classes.real_nodes import PythonNode, RealNode, as_node
from ._body_templates import InteractiveBodyTemplate
from ._node_templates import NodeTemplate
from ._decorators import (DeltaBlock,
                          DeltaMethodBlock,
                          Interactive)
from ._delta_graph import DeltaGraph
from ._placeholder_factory import placeholder_node_factory


# user-facing classes
__all__ = ["DeltaBlock",
           "DeltaMethodBlock",
           "Interactive",
           "DeltaGraph",
           "RealNode",
           "PythonNode",
           "MigenNodeTemplate",
           "placeholder_node_factory",
           "NodeTemplate"]



"""Classes to represent different node bodies a Deltaflow node could represent.
"""
from abc import ABC, abstractmethod
from copy import copy
import dill
from typing import Any, Callable, List

from .latency import Latency


class Body(ABC):

    def __init__(self, latency: Latency = Latency(time=300), tags: List[str] = []):
        """
        Parameters
        ----------
        latency : Latency
            Estimated time this body takes to run
        tags : List[str]
            List of strings to be added as access tags.
        """
        self._access_tags = tags + type(self).mro()
        self.latency = latency

    @property
    def access_tags(self) -> List[object]:
        """List of ways this body wants to identify itself.
        These are usually a list of classes and a list of user added strings.

        Returns
        -------
        List[object] 
            A list of immutable objects to be used as keys in a dict
        """
        return self._access_tags

    @property
    @abstractmethod
    def language(self) -> str:
        """String representation of the language represented by this body class.

        Returns
        -------
        str
        """
        pass

    @property
    @abstractmethod
    def as_serialized(self) -> str:
        """Serialisation of body as string.

        Returns
        -------
        str
        """
        pass


class PythonBody(Body):

    @property
    def language(self) -> str:
        """Language of the body.

        Returns
        -------
        str
        """
        return "Py"

    @property
    def as_serialized(self) -> str:
        """Serialisation of body as string.

        Returns
        -------
        str
        """
        return dill.dumps(self, recurse=True)


class PyFuncBody(PythonBody):
    """Node body for python functions.

    Parameters
    ----------
    fn : Callable
        A callable object that will play a role of the body.
    environment : Callable
        Link to the environment.
    """

    def __init__(self, fn: Callable,
                 latency: Latency = Latency(time=350),
                 tags: List[str] = []):
        super().__init__(latency, tags + [fn.__name__])
        self.callback = fn

    def eval(self, *args, **kwargs):
        return self.callback(*args, **kwargs)


class PyConstBody(PythonBody):
    """Node body for python are to be treated as constant.

    This means when its value is first queried it evaluates and from then on
    always returns this value.

    Parameters
    ----------
    fn : Callable
        A callable object that will play a role of the body.
    environment : Callable
        Link to the environment.
    args
        Constant arguments for the body.
    kwargs
        Constant keyworded arguments for the body.
    """

    def __init__(self, fn, *args,
                 latency: Latency = Latency(time=100),
                 tags: List[str] = [], **kwargs):
        super().__init__(latency, tags + [fn.__name__])
        self.value = None
        self.callback = fn
        self.args = args
        self.kwargs = kwargs

    def eval(self):
        if self.value is None:
            # Evaluate value the first time the node is evaluated
            evaluated_args = []
            for arg in self.args:
                possible_arg = getattr(arg, "body", None)
                if possible_arg is not None:
                    evaluated_arg = possible_arg.eval()
                    # If previous node has forked output
                    # only get the relevant argument
                    possible_index = getattr(arg, "index", None)
                    if possible_index is not None:
                        evaluated_arg = getattr(
                            evaluated_arg, possible_index, None)
                    if evaluated_arg is not None:
                        evaluated_args.append(evaluated_arg)
                    else:
                        raise ValueError("Node with a constant body cannot "
                                         "receive \'None\' as input.")

            evaluated_kwargs = {}
            for key, arg in self.kwargs.items():
                possible_kwarg = getattr(arg, "body", None)
                if possible_kwarg is not None:
                    evaluated_kwarg = possible_kwarg.eval()
                    # If previous node has forked output
                    # only get the relevant keyword argument
                    possible_index = getattr(arg, "index", None)
                    if possible_index is not None:
                        evaluated_kwarg = getattr(
                            evaluated_kwarg, possible_index, None)
                    if evaluated_kwarg is not None:
                        evaluated_kwargs[key] = evaluated_kwarg
                    else:
                        raise ValueError("Node with a constant body cannot "
                                         "receive \'None\' as input.")

            self.value = self.callback(*evaluated_args, **evaluated_kwargs)

        return self.value


class PyMethodBody(PythonBody):
    """Node body for python methods.

    Must store instance to run with when instantiated. Uses this instance when
    evaluated.

    Parameters
    ----------
    fn : Callable
        A callable object that will play a role of the body.
    instance : object
        Object instance for state storage.
    environment : Callable
        Link to the environment.
    """

    def __init__(self, fn, instance,
                 latency: Latency = Latency(time=350),
                 tags: List[str] = []):
        super().__init__(latency, tags + [fn.__name__])
        self.callback = fn
        self.instance = instance

    def eval(self, *args, **kwargs):
        return self.callback(self.instance, *args, **kwargs)


class PyMigenBody(PyMethodBody):
    """Node body for migen methods.

    ..
        It's basically :py:class:`PyMethodBody`, but...
        the node is evaluated multiple times until we have an output.

        The reason for this is the following:

        - if a migen node has multiple outputs, they will be forked using
        ``make_forked_return``
        - the result of the node is ready only after several cycles
        - in the meantime the result should be set to _some_ value, like
          ``None`` or ``ForkedReturn(None, None, ...)``
        - I chose to set the entire output to None until the node produces
        a result

        This policy should be revised when we start working on protocol
        adaptors.

        :py:meth:`MigenNodeTemplate.tb_generator` has the number of iterations
        defined, that number should be chosen such that it's enough to produce
        an output. Otherwise this body will loop forever.
    """

    def __init__(self, fn, instance,
                 latency: Latency = Latency(clocks=1),
                 tags: List[str] = []):
        super().__init__(fn, instance, latency, tags)

    def eval(self, *args, **kwargs):
        """Overwrites :meth:`PyMethodBody.eval`."""
        while True:
            ret = self.callback(self.instance, *args, **kwargs)
            if ret is not None:
                return ret

    @property
    def as_serialized(self) -> str:
        """Serialisation of body as string.

        Returns
        -------
        str
        """
        return str(self.instance.get_serialised_body())


class PyInteractiveBody(PyFuncBody):
    """Body class to represent bodies that expose queues to the designer.
    We explicitily define this class to enable custom-code creation in the
    runtimes.
    """
